// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include <schpin_lib/logging.h>
//
#include "fw/reactor.h"
#include "fw/serial.h"

#include <chrono>
#include <schpin_lib_exp/config.h>
#include <schpin_msgs/msg/control_cmd.hpp>

using namespace std::chrono_literals;
using namespace schpin;

struct control_pc_config
{
        std::string device_filename;

        static auto ros_param_get()
        {
                return std::make_tuple( ros_opt_param< std::string >{
                    "device", "device to open for serial communication", "/dev/ttyUSB0" } );
        }
};

class control_node
{

        control_pc_config config_;
        fw_reactor< 20 >  reactor_;
        serial_connection ser_;

        rclcpp::Publisher< schpin_msgs::msg::ControlCmd >::SharedPtr pub_;

public:
        control_node( const control_node& ) = delete;
        control_node( control_node&& )      = delete;
        control_node& operator=( const control_node& ) = delete;
        control_node& operator=( control_node&& ) = delete;

        control_node( rclcpp::Node& rcnode, serial_connection ser, control_pc_config config )
          : config_( std::move( config ) )
          , ser_( ser )
          , pub_( rcnode.create_publisher< schpin_msgs::msg::ControlCmd >(
                "/control/feedback",
                200 ) )
        {
                // TODO no delay subscription?
        }

        auto make_matcher( ctl_id id = ctl_id{ 0 } )
        {
                return match(
                    [this, id]( control_tag, fw_node_group_message group_msg ) {
                            auto m = fw_node_protocol::serialize( id, group_msg, checksum{} );

                            schpin_msgs::msg::ControlCmd msg;
                            msg.data = map_f_to_v( m, []( uint8_t val ) {
                                    return val;
                            } );
                            pub_->publish( msg );
                    },
                    [this]( servo_tag, s_ctl_id, reply, const master_dynmx_message& ) {},
                    [this]( servo_tag, s_ctl_id msg_id, reply r, const master_lewan_message& m ) {
                            ser_.write( m );
                            if ( r == reply::YES ) {
                                    receive_serial_msg( msg_id, m );
                            }
                    } );
        }
        void receive_serial_msg( s_ctl_id msg_id, const master_lewan_message& m )
        {
                static constexpr std::size_t n = lewan_master_message::max_size;
                ser_.read< n >().match(
                    [&]( const em::static_vector< uint8_t, n >& data ) {
                            lewan_sequencer< lewan_master_message > sequencer;

                            bool load_fired = sequencer.load( data, [&]( lewan_master_message m ) {
                                    reactor_.on_servo_receive(
                                        get_system_time(), msg_id, m, make_matcher() );
                            } );

                            if ( !load_fired ) {
                                    reactor_.on_servo_not_replying( m, make_matcher() );
                            }
                    },
                    [&]( auto err ) {
                            reactor_.on_servo_not_replying( m, make_matcher() );
                            SCHPIN_ERROR_LOG( "control", err );
                    } );
        }

        void control_topic( schpin_msgs::msg::ControlCmd::SharedPtr msg )
        {
                auto opt_msg = node_fw_message::make( msg->data );
                if ( !opt_msg ) {
                        SCHPIN_ERROR_LOG(
                            "node", "Data that arrived from control topic are too long" );
                        return;
                }
                parse_node_fw( *opt_msg )
                    .match(
                        em::uncurry |
                            [&]( ctl_id id, node_fw_group_variant submsg, checksum ) {
                                    reactor_.on_control_receive( id, submsg, make_matcher( id ) );
                            },
                        [&]( auto ) {  // TODO: print the error
                                SCHPIN_ERROR_LOG( "control_pc", "protocol parsin error" );
                        } );
        }

        void tick( clk_time clock_time )
        {
                reactor_.tick( clock_time, make_matcher() );
        }

        ~control_node() = default;
};

class control_supervizor : public rclcpp::Node
{
        std::optional< control_node >                                   control_;
        rclcpp::TimerBase::SharedPtr                                    timer_;
        rclcpp::Subscription< schpin_msgs::msg::ControlCmd >::SharedPtr sub_;

public:
        control_supervizor()
          : Node( "control" )
          , sub_( this->create_subscription< schpin_msgs::msg::ControlCmd >(
                "/control/control",
                200,
                std::bind( &control_supervizor::control_topic, this, std::placeholders::_1 ) ) )
        {
                get_config< control_pc_config >( *this )
                    .bind_left( [&]( control_pc_config config ) {  //
                            return serial_connection::make( config.device_filename )
                                .convert_left( [&]( serial_connection port ) {
                                        return std::make_pair( port, config );
                                } );
                    } )
                    .match(
                        em::uncurry |
                            [&]( serial_connection port, control_pc_config conf ) {
                                    control_.emplace( *this, port, std::move( conf ) );

                                    SCHPIN_INFO_LOG( "node", "Node initialized and running" );

                                    timer_ = this->create_wall_timer(
                                        1ms, std::bind( &control_supervizor::tick, this ) );
                            },
                        [&]( auto e ) {  //
                                std::cerr << e;
                        } );
        }

        void control_topic( schpin_msgs::msg::ControlCmd::SharedPtr msg )
        {
                control_->control_topic( msg );
        }

        void tick()
        {
                control_->tick( get_system_time() );
        }
};

int main( int argc, char* argv[] )
{
        rclcpp::init( argc, argv );
        rclcpp::spin( std::make_shared< control_supervizor >() );
        rclcpp::shutdown();
        return 0;
}
