// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include <schpin_lib_exp/bounded.h>
#include <schpin_lib_exp/protocol/item.h>
#include <schpin_lib_exp/protocol/offset.h>
#include <schpin_lib_exp/protocol/sizeless_message.h>

#pragma once

namespace schpin
{

struct dynmx_error
{
        // TODO make this enum
        bool overload;
        bool electrical_shock;
        bool motor_encoder;
        bool overheating;
        bool input_voltage;
};

inline std::ostream& operator<<( std::ostream& os, dynmx_error err )
{
        std::string s = "";
        if ( err.overload ) {
                os << "OVERLOAD";
                s = ",";
        }
        if ( err.electrical_shock ) {
                os << s << "ESHOCK";
                s = ",";
        }
        if ( err.motor_encoder ) {
                os << s << "M_ENCOD";
                s = ",";
        }
        if ( err.overheating ) {
                os << s << "HEAT";
                s = ",";
        }
        if ( err.input_voltage ) {
                os << s << "IVOT";
        }
        return os;
}

struct dynmx_status
{
        enum profile_enum
        {
                TRAPEZOIADL_PROFILE = 0b11,
                TRIANGULAR_PROFILE  = 0b10,
                RECTANGULAR_PROFILE = 0b01,
                NO_PROFILE          = 0b00
        };

        profile_enum profile;
        bool         following;
        bool         ongoing;
        bool         arrived;
};
inline std::ostream& operator<<( std::ostream& os, dynmx_status st )
{
        os << magic_enum::enum_name( st.profile );
        if ( st.following ) {
                os << ",FOLLOWING";
        }
        if ( st.ongoing ) {
                os << ",ONGOING";
        }
        if ( st.arrived ) {
                os << ",ARRIVED";
        }
        return os;
}

using dynmx_header                           = bounded< uint32_t, 0xFFFFFD00, 0xFFFFFD00 >;
static constexpr dynmx_header dynmx_header_v = dynmx_header::get< 0xFFFFFD00 >();
using dynmx_servo_id                         = bounded< uint8_t, 0, 252 >;
using dynmx_packet_size    = protocol_offset< struct dynmx_packet_size_tag, uint16_t, 2 >;
using dynmx_crc            = em::tagged_quantity< struct dynmx_crc_tag, uint16_t >;
using dynmx_response_error = bounded< uint8_t, 0x0, 0x7 >;  // maybe enum here
static constexpr dynmx_response_error dynmx_no_error         = dynmx_response_error::get< 0x00 >();
static constexpr dynmx_response_error dynmx_result_error     = dynmx_response_error::get< 0x01 >();
static constexpr dynmx_response_error dynmx_instru_error     = dynmx_response_error::get< 0x02 >();
static constexpr dynmx_response_error dynmx_crc_error        = dynmx_response_error::get< 0x03 >();
static constexpr dynmx_response_error dynmx_data_range_error = dynmx_response_error::get< 0x04 >();
static constexpr dynmx_response_error dynmx_data_len_error   = dynmx_response_error::get< 0x05 >();
static constexpr dynmx_response_error dynmx_data_lim_error   = dynmx_response_error::get< 0x06 >();
static constexpr dynmx_response_error dynmx_access_error     = dynmx_response_error::get< 0x07 >();
using dynmx_model_number = em::tagged_quantity< struct dynmx_model_number_tag, uint16_t >;
using dynmx_model_info   = em::tagged_quantity< struct dynmx_model_info_tag, uint16_t >;
using dynmx_fw_version   = em::tagged_quantity< struct dynmx_fw_version_tag, uint8_t >;
using dynmx_baud_rate    = bounded< uint8_t, 0, 7 >;
using dynmx_delay_time   = bounded< uint8_t, 0, 254 >;
using dynmx_drive_mode   = bounded< uint8_t, 0, 5 >;
using dynmx_oper_mode    = bounded< uint8_t, 0, 16 >;
static constexpr auto dynmx_velocity_control_mode          = dynmx_oper_mode::get< 1 >();
static constexpr auto dynmx_position_control_mode          = dynmx_oper_mode::get< 3 >();
static constexpr auto dynmx_extended_position_control_mode = dynmx_oper_mode::get< 4 >();
static constexpr auto dynmx_pwm_control_mode               = dynmx_oper_mode::get< 16 >();
using dynmx_proto_type                                     = bounded< uint8_t, 1, 2 >;
using dynmx_home_offset           = bounded< int32_t, -1'044'479, 1'044'479 >;
using dynmx_vel_limit             = bounded< uint32_t, 0, 1'023 >;
using dynmx_voltage               = bounded< uint16_t, 60, 140 >;
using dynmx_temp                  = bounded< uint8_t, 0, 100 >;
using dynmx_pwm_limit             = bounded< uint16_t, 0, 885 >;
using dynmx_pos                   = bounded< uint32_t, 0, 4'095 >;
using dynmx_angle                 = dynmx_pos;
using dynmx_bool                  = bounded< uint8_t, 0, 1 >;
static constexpr auto dynmx_true  = dynmx_bool::get< 1 >();
static constexpr auto dynmx_false = dynmx_bool::get< 0 >();
using dynmx_status_level          = bounded< uint8_t, 0, 2 >;
using dynmx_pid_gain              = bounded< uint16_t, 0, 16'383 >;
using dynmx_watchdog              = bounded< int16_t, -1, 127 >;
using dynmx_pwm                   = bounded< int16_t, -885, 885 >;
using dynmx_vel                   = bounded< int32_t, -1'023, 1'023 >;
using dynmx_angular_velocity      = dynmx_vel;
using dynmx_profile_accel         = bounded< uint32_t, 0, 32'737 >;
using dynmx_profile_vel           = bounded< uint32_t, 0, 32'767 >;
using dynmx_time                  = bounded< uint16_t, 0, 32'767 >;
using dynmx_load                  = bounded< int16_t, -1'000, 1'000 >;
using dynmx_addr                  = em::tagged_quantity< struct dynmx_addr_tag, uint16_t >;
using dynmx_len                   = em::tagged_quantity< struct dynmx_len_tag, uint16_t >;
using dynmx_buffer                = prot_sizeless_message< 16 >;
using dynmx_reset_flags = bounded< uint8_t, 0x01, 0x02 >;  // missing 0xFF option, maybe new type?

template <>
inline constexpr endianess_enum serialization_endianess< dynmx_packet_size > = endianess_enum::BIG;
template <>
inline constexpr endianess_enum serialization_endianess< dynmx_addr > = endianess_enum::BIG;
template <>
inline constexpr endianess_enum serialization_endianess< dynmx_len > = endianess_enum::BIG;
template <>
inline constexpr endianess_enum serialization_endianess< dynmx_model_number > = endianess_enum::BIG;
template <>
inline constexpr endianess_enum serialization_endianess< dynmx_model_info > = endianess_enum::BIG;
template <>
inline constexpr endianess_enum serialization_endianess< dynmx_home_offset > = endianess_enum::BIG;
template <>
inline constexpr endianess_enum serialization_endianess< dynmx_vel_limit > = endianess_enum::BIG;
template <>
inline constexpr endianess_enum serialization_endianess< dynmx_voltage > = endianess_enum::BIG;
template <>
inline constexpr endianess_enum serialization_endianess< dynmx_pwm_limit > = endianess_enum::BIG;
template <>
inline constexpr endianess_enum serialization_endianess< dynmx_pos > = endianess_enum::BIG;
template <>
inline constexpr endianess_enum serialization_endianess< dynmx_pid_gain > = endianess_enum::BIG;
template <>
inline constexpr endianess_enum serialization_endianess< dynmx_watchdog > = endianess_enum::BIG;
template <>
inline constexpr endianess_enum serialization_endianess< dynmx_pwm > = endianess_enum::BIG;
template <>
inline constexpr endianess_enum serialization_endianess< dynmx_vel > = endianess_enum::BIG;
template <>
inline constexpr endianess_enum serialization_endianess< dynmx_profile_accel > =
    endianess_enum::BIG;
template <>
inline constexpr endianess_enum serialization_endianess< dynmx_profile_vel > = endianess_enum::BIG;
template <>
inline constexpr endianess_enum serialization_endianess< dynmx_time > = endianess_enum::BIG;
// redefinition :/ TODO need to tag bounded vals
// template <>
// inline constexpr endianess_enum serialization_endianess< dynmx_load > = endianess_enum::BIG;

template <>
struct protocol_item< dynmx_error > : protocol_item_base< dynmx_error >
{
        using input_type  = dynmx_error;
        using output_type = dynmx_error;

        static constexpr std::size_t size = 1;

        static std::size_t
        serialize_at( static_view< size, uint8_t* > buffer, const dynmx_error& e )
        {
                std::bitset< 8 > bset;
                bset[5] = e.overload;
                bset[4] = e.electrical_shock;
                bset[3] = e.motor_encoder;
                bset[2] = e.overheating;
                bset[0] = e.input_voltage;
                return protocol_item< std::bitset< 8 > >::serialize_at( buffer, bset );
        }
        template < typename Buffer, typename Visitor >
        static constexpr std::size_t deserialize( Buffer&& buffer, Visitor&& vis )
        {
                return protocol_item< std::bitset< 8 > >::deserialize(
                    buffer,
                    match(
                        [&]( value_tag, std::bitset< 8 > bset ) {
                                dynmx_error de{};
                                de.overload         = bset[5];
                                de.electrical_shock = bset[4];
                                de.motor_encoder    = bset[3];
                                de.overheating      = bset[2];
                                de.input_voltage    = bset[0];
                                vis( value_tag{}, de );
                        },
                        [&]( error_tag, auto e ) {
                                vis( error_tag{}, e );
                        } ) );
        }
};

template <>
struct protocol_item< dynmx_status > : protocol_item_base< dynmx_status >
{
        using input_type  = dynmx_status;
        using output_type = dynmx_status;

        static constexpr std::size_t size = 1;

        using sub_prot = protocol_item< std::bitset< 8 > >;

        static std::size_t
        serialize_at( static_view< size, uint8_t* > buffer, const dynmx_status& st )
        {
                std::bitset< 8 > bset;
                bset[5] = st.profile;
                bset[4] = st.profile >> 1;
                bset[3] = st.following;
                bset[1] = st.ongoing;
                bset[0] = st.arrived;
                return sub_prot::serialize_at( buffer, bset );
        }

        template < typename Buffer, typename Visitor >
        static constexpr std::size_t deserialize( Buffer&& buffer, Visitor&& vis )
        {
                return sub_prot::deserialize(
                    buffer,
                    match(
                        [&]( value_tag, std::bitset< 8 > bset ) {
                                dynmx_status st;
                                st.profile = static_cast< dynmx_status::profile_enum >(
                                    bset[4] << 1 & bset[5] );
                                st.following = bset[3];
                                st.ongoing   = bset[1];
                                st.arrived   = bset[0];
                                vis( value_tag{}, st );
                        },
                        [&]( error_tag, auto e ) {
                                vis( error_tag{}, e );
                        } ) );
        }
};

}  // namespace schpin
