// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "fw/dynmx/config.h"
#include "fw/dynmx/context.h"
#include "fw/dynmx/conv.h"
#include "fw/servo/base.h"
#include "shared/protocol.h"

#pragma once

namespace schpin
{

struct dynmx_goto_event
{
        dynmx_angle goal_pos;
        btime       t;
};

struct dynmx_spin_event
{
        dynmx_angular_velocity velocity;
        btime                  t;
};

struct dynmx_config_event
{
        dynmx_config_enum reg;
        config_buffer     buffer;
};

using dynmx_servo_events_variant =
    std::variant< dynmx_goto_event, dynmx_spin_event, dynmx_config_event >;

class dynmx_servo
{
        struct idle_task
        {
        };

        struct activate_task
        {
                std::size_t i;
        };

        struct goto_task
        {
                dynmx_angle from_pos;
                dynmx_angle to_pos;
                clk_time    from_time;
                clk_time    to_time;
        };

        struct spin_task
        {
        };

        using task_variant = std::variant< idle_task, activate_task, goto_task, spin_task >;

        dynmx_config_map config_map_;
        dynmx_context    context_;
        dynmx_converter  converter_;
        task_variant     task_ = { activate_task{} };

        std::size_t poll_i_;

public:
        using receive_type   = dynmx_buffer;
        using events_variant = dynmx_servo_events_variant;

        explicit dynmx_servo( dynmx_servo_id id )
          : context_( id )
        {
        }

        servo_id get_id() const
        {
                return *context_.get_id();
        }

        bool has_task() const
        {
                return !std::holds_alternative< idle_task >( task_ );
        }

        bool is_activating() const
        {
                return std::holds_alternative< activate_task >( task_ );
        }

        template < typename Visitor >
        void stop( Visitor&& )
        {
        }

        std::optional< servo_master_message >
        on_servo_receive( clk_time, s_ctl_id sid, const dynmx_buffer& buff )
        {

                context_.on_receive( sid, buff );

                return {};
        }

        template < typename Visitor >
        void tick( clk_time now, Visitor&& vis )
        {
                task_ << match(
                    [&]( idle_task ) {},
                    [&]( activate_task at ) {
                            at.i = context_.full_poll( at.i, vis );
                            if ( at.i == 0 ) {
                                    task_ = idle_task{};
                            } else {
                                    task_ = at;
                            }
                    },
                    [&]( goto_task gt ) {
                            clk_time t_time = em::clamp( now, gt.from_time, gt.to_time );
                            float    mapped = map_val< float >(
                                t_time, gt.from_time, gt.to_time, *gt.from_pos, *gt.to_pos );

                            context_.set_item< DYNMX_GOAL_POS >(
                                *dynmx_angle::make( mapped ), vis );

                            SCHPIN_INFO_LOG( "dynmx", "goto: " << mapped );

                            if ( now < gt.to_time ) {
                                    return;
                            }

                            if ( abs( static_cast< int >(
                                          *context_.get_item< DYNMX_PRESENT_POS >() ) -
                                      static_cast< int >( *gt.to_pos ) ) > 15 ) {
                                    return;
                            }

                            task_ = idle_task{};
                    },
                    [&]( spin_task ) {} );
        }

        template < typename Visitor >
        void on_event( ctl_id, clk_time now, dynmx_servo_events_variant var, Visitor&& vis )
        {
                var << match(
                    [&]( dynmx_goto_event e ) {
                            context_.set_item< DYNMX_TORQUE_EN >( dynmx_false, vis );
                            context_.set_item< DYNMX_OPER_MODE >(
                                dynmx_position_control_mode, vis );
                            context_.set_item< DYNMX_TORQUE_EN >( dynmx_true, vis );

                            task_ = goto_task{
                                context_.get_item< DYNMX_PRESENT_POS >(),
                                e.goal_pos,
                                now,
                                now + clk_time{ *e.t } };
                    },
                    [&]( dynmx_spin_event ) {},
                    [&]( dynmx_config_event ) {} );
        }

        template < typename Visitor >
        void poll( clk_time, std::size_t i_offset, Visitor&& vis )
        {
                poll_i_       = ( poll_i_ + 1 ) % 13;
                std::size_t i = ( i_offset + poll_i_ ) % 13;
                switch ( i ) {
                        case 4:
                                context_.query_item< DYNMX_PRESENT_VOLTAGE >( vis );
                                break;
                        case 8:
                                context_.query_item< DYNMX_PRESENT_TEMP >( vis );
                                break;
                        default:
                                context_.query_item< DYNMX_PRESENT_POS >( vis );
                                break;
                }
        }

        em::either< dynmx_servo_events_variant, servo_master_message >
        on_servo_packet( ctl_id cid, master_servo_command_variant var )
        {
                using return_type = em::either< dynmx_servo_events_variant, servo_master_message >;

                return var << tuple_match(
                           [&]( cmd_tag< GOTO >, bangle a, btime t ) -> return_type {
                                   return handle_in_to_se( converter_, cid, GOTO, a )
                                       .convert_left(
                                           [&]( dynmx_angle an ) -> dynmx_servo_events_variant {
                                                   return dynmx_goto_event{ an, t };
                                           } );
                           },
                           [&]( cmd_tag< SPIN >, bangular_velocity a, btime t ) -> return_type {
                                   return handle_in_to_se( converter_, cid, SPIN, a )
                                       .convert_left(
                                           [&]( dynmx_angular_velocity av )
                                               -> dynmx_servo_events_variant {
                                                   return dynmx_spin_event{ av, t };
                                           } );
                           },
                           [&]( cmd_tag< POSITION > ) -> return_type {
                                   return make_servo_message< POSITION >( converter_.se_to_in(
                                       context_.get_item< DYNMX_PRESENT_POS >() ) );
                           },
                           [&]( cmd_tag< TEMP > ) -> return_type {
                                   return make_servo_message< TEMP >( converter_.se_to_in(
                                       context_.get_item< DYNMX_PRESENT_TEMP >() ) );
                           },
                           [&]( cmd_tag< VOLTAGE > ) -> return_type {
                                   return make_servo_message< VOLTAGE >( converter_.se_to_in(
                                       context_.get_item< DYNMX_PRESENT_VOLTAGE >() ) );
                           },
                           [&]( cmd_tag< VELOCITY > ) -> return_type {
                                   return make_servo_message< VELOCITY >( converter_.se_to_in(
                                       context_.get_item< DYNMX_PRESENT_VEL >() ) );
                           },
                           [&]( cmd_tag< TASK > ) -> return_type {
                                   servo_task_enum e = task_ << match(
                                                           [&]( idle_task ) {
                                                                   return TASK_IDLE;
                                                           },
                                                           [&]( activate_task ) {
                                                                   return TASK_ACTIVATE;
                                                           },
                                                           [&]( goto_task ) {
                                                                   return TASK_GOTO;
                                                           },
                                                           [&]( spin_task ) {
                                                                   return TASK_SPIN;
                                                           } );

                                   return make_servo_message< TASK >( e );
                           },
                           [&]( cmd_tag< CONFIG_GET >, config_addr addr ) -> return_type {
                                   return make_servo_message< CONFIG_GET >(
                                       addr,
                                       config_map_.get_buffered_val< config_buffer >(
                                           static_cast< dynmx_config_enum >( addr ) ) );
                           },
                           [&]( cmd_tag< CONFIG_SET >,
                                config_addr   addr,
                                config_buffer buffr ) -> return_type {
                                   return dynmx_servo_events_variant{ dynmx_config_event{
                                       static_cast< dynmx_config_enum >( addr ), buffr } };
                           } );
        }

        void dump_info()
        {
                context_.dump_info();
        }
};
}  // namespace schpin
