// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "fw/lewan/base.h"

#include <emlabcpp/static_circular_buffer.h>
#include <schpin_lib_exp/protocol.h>

#pragma once

namespace schpin
{
enum lewan_message_enum : uint8_t
{
        LEWAN_SERVO_MOVE_TIME_WRITE      = 1,
        LEWAN_SERVO_MOVE_TIME_READ       = 2,
        LEWAN_SERVO_MOVE_TIME_WAIT_WRITE = 7,
        LEWAN_SERVO_MOVE_TIME_WAIT_READ  = 8,
        LEWAN_SERVO_MOVE_START           = 11,
        LEWAN_SERVO_MOVE_STOP            = 12,
        LEWAN_SERVO_ID_WRITE             = 13,
        LEWAN_SERVO_ID_READ              = 14,
        LEWAN_SERVO_ANGLE_OFFSET_ADJUST  = 17,
        LEWAN_SERVO_ANGLE_OFFSET_WRITE   = 18,
        LEWAN_SERVO_ANGLE_OFFSET_READ    = 19,
        LEWAN_SERVO_ANGLE_LIMIT_WRITE    = 20,
        LEWAN_SERVO_ANGLE_LIMIT_READ     = 21,
        LEWAN_SERVO_VIN_LIMIT_WRITE      = 22,
        LEWAN_SERVO_VIN_LIMIT_READ       = 23,
        LEWAN_SERVO_TEMP_MAX_LIMIT_WRITE = 24,
        LEWAN_SERVO_TEMP_MAX_LIMIT_READ  = 25,
        LEWAN_SERVO_TEMP_READ            = 26,
        LEWAN_SERVO_VIN_READ             = 27,
        LEWAN_SERVO_POS_READ             = 28,
        LEWAN_SERVO_OR_MOTOR_MODE_WRITE  = 29,
        LEWAN_SERVO_OR_MOTOR_MODE_READ   = 30,
        LEWAN_SERVO_LOAD_OR_UNLOAD_WRITE = 31,
        LEWAN_SERVO_LOAD_OR_UNLOAD_READ  = 32,
        LEWAN_SERVO_LED_CTRL_WRITE       = 33,
        LEWAN_SERVO_LED_CTRL_READ        = 34,
        LEWAN_SERVO_LED_ERROR_WRITE      = 35,
        LEWAN_SERVO_LED_ERROR_READ       = 36
};

struct master_lewan_group
  : command_group<>::with_msg_counter< lewan_packet_size >::with_commands<
        command< LEWAN_SERVO_MOVE_TIME_WRITE >::with_args< lewan_angle, lewan_time >,
        command< LEWAN_SERVO_MOVE_TIME_READ >,
        command< LEWAN_SERVO_MOVE_TIME_WAIT_WRITE >::with_args< lewan_angle, lewan_time >,
        command< LEWAN_SERVO_MOVE_TIME_WAIT_READ >,
        command< LEWAN_SERVO_MOVE_START >,
        command< LEWAN_SERVO_MOVE_STOP >,
        command< LEWAN_SERVO_ID_WRITE >::with_args< lewan_servo_id >,
        command< LEWAN_SERVO_ID_READ >,
        command< LEWAN_SERVO_ANGLE_OFFSET_ADJUST >::with_args< lewan_angle_offset >,
        command< LEWAN_SERVO_ANGLE_OFFSET_WRITE >,
        command< LEWAN_SERVO_ANGLE_OFFSET_READ >,
        command< LEWAN_SERVO_ANGLE_LIMIT_WRITE >::with_args< lewan_angle, lewan_angle >,
        command< LEWAN_SERVO_ANGLE_LIMIT_READ >,
        command< LEWAN_SERVO_VIN_LIMIT_WRITE >::with_args< lewan_voltage, lewan_voltage >,
        command< LEWAN_SERVO_VIN_LIMIT_READ >,
        command< LEWAN_SERVO_TEMP_MAX_LIMIT_WRITE >::with_args< lewan_temp >,
        command< LEWAN_SERVO_TEMP_MAX_LIMIT_READ >,
        command< LEWAN_SERVO_TEMP_READ >,
        command< LEWAN_SERVO_VIN_READ >,
        command< LEWAN_SERVO_POS_READ >,
        command< LEWAN_SERVO_OR_MOTOR_MODE_WRITE >::
            with_args< lewan_servo_mode, uint8_t, lewan_angular_velocity >,
        command< LEWAN_SERVO_OR_MOTOR_MODE_READ >,
        command< LEWAN_SERVO_LOAD_OR_UNLOAD_WRITE >::with_args< lewan_load >,
        command< LEWAN_SERVO_LOAD_OR_UNLOAD_READ >,
        command< LEWAN_SERVO_LED_CTRL_WRITE >::with_args< lewan_led_mode >,
        command< LEWAN_SERVO_LED_CTRL_READ >,
        command< LEWAN_SERVO_LED_ERROR_WRITE >::with_args< lewan_led_error >,
        command< LEWAN_SERVO_LED_ERROR_READ > >
{
};

using master_lewan_submessage = master_lewan_group::message_t;
using master_lewan_variant    = master_lewan_group::command_variant;

struct lewan_master_group
  : command_group<>::with_msg_counter< lewan_packet_size >::with_commands<
        command< LEWAN_SERVO_MOVE_TIME_READ >::with_args< lewan_angle, lewan_time >,
        command< LEWAN_SERVO_MOVE_TIME_WAIT_READ >::with_args< lewan_angle, lewan_time >,
        command< LEWAN_SERVO_ID_READ >::with_args< lewan_servo_id >,
        command< LEWAN_SERVO_ANGLE_OFFSET_READ >::with_args< lewan_angle_offset >,
        command< LEWAN_SERVO_ANGLE_LIMIT_READ >::with_args< lewan_angle, lewan_angle >,
        command< LEWAN_SERVO_VIN_LIMIT_READ >::with_args< lewan_voltage, lewan_voltage >,
        command< LEWAN_SERVO_TEMP_MAX_LIMIT_READ >::with_args< lewan_temp >,
        command< LEWAN_SERVO_TEMP_READ >::with_args< lewan_temp >,
        command< LEWAN_SERVO_VIN_READ >::with_args< lewan_voltage >,
        command< LEWAN_SERVO_POS_READ >::with_args< lewan_full_angle >,
        command< LEWAN_SERVO_OR_MOTOR_MODE_READ >::
            with_args< lewan_servo_mode, uint8_t, lewan_angular_velocity >,
        command< LEWAN_SERVO_LOAD_OR_UNLOAD_READ >::with_args< lewan_load >,
        command< LEWAN_SERVO_LED_CTRL_READ >::with_args< lewan_led_mode >,
        command< LEWAN_SERVO_LED_ERROR_READ >::with_args< lewan_led_error > >
{
};

using lewan_master_submessage = lewan_master_group::message_t;
using lewan_master_variant    = lewan_master_group::command_variant;

template < typename Message >
class lewan_sequencer
{
        // soo, looking for 2x lewan_header_start , anything, lewan_packet_size

        enum states
        {
                HEADER_START_1,
                HEADER_START_2,
                SERVO_ID,
                PACKET_SIZE,
                PACKET_DATA
        };

        em::static_circular_buffer< uint8_t, 64 > buffer_;

        states      state_       = HEADER_START_1;
        std::size_t start_       = 0;
        std::size_t msg_size_    = 0;
        std::size_t ideal_batch_ = 2;

public:
        template < typename Data, typename MsgCallback >
        bool load( const Data& data, MsgCallback&& cb )
        {
                auto data_view = em::view{ data };

                if ( data_view.empty() ) {
                        ideal_batch_ = 2;
                        return false;
                }
                buffer_.push_back( data_view.front() );
                bool has_fired = false;
                switch ( state_ ) {
                        case HEADER_START_1:
                                if ( data_view.front() == *lewan_header_start_v ) {
                                        state_ = HEADER_START_2;
                                } else {
                                        buffer_.clear();
                                }
                                break;
                        case HEADER_START_2:
                                if ( data_view.front() == *lewan_header_start_v ) {
                                        state_ = SERVO_ID;
                                } else {
                                        state_ = HEADER_START_1;
                                        buffer_.clear();
                                }
                                break;
                        case SERVO_ID:
                                state_ = PACKET_SIZE;
                                break;
                        case PACKET_SIZE:
                                msg_size_ = static_cast< std::size_t >( data_view.front() - 2 );
                                state_    = PACKET_DATA;
                                break;
                        case PACKET_DATA:
                                if ( msg_size_ != 0 ) {
                                        msg_size_ -= 1;
                                        break;
                                }
                                auto opt_msg = Message::make( buffer_ );
                                if ( opt_msg ) {
                                        has_fired = true;
                                        cb( *opt_msg );
                                } else {
                                        // ERROR
                                }

                                buffer_.clear();
                                state_ = HEADER_START_1;
                                break;
                }
                return load( em::tail( data_view ), cb ) || has_fired;
        }
};

template < typename Msg >
inline uint8_t make_lewan_checksum( const Msg& msg )
{
        unsigned checksum =
            em::sum( em::view{ msg.begin() + 2, msg.end() - 1 }, em::convert_to< unsigned >{} );

        return static_cast< uint8_t >( ~( checksum & 0xff ) );
}

template < typename T >
using lewan_protocol_tuple = protocol_tuple<>::with_items<
    lewan_header_start,
    lewan_header_start,
    lewan_servo_id,
    protocol_recursive_promoter< T >,
    lewan_checksum >;

using lewan_master_protocol = lewan_protocol_tuple< lewan_master_submessage >;
using lewan_master_message  = lewan_master_protocol::message_t;

using master_lewan_protocol = lewan_protocol_tuple< master_lewan_submessage >;
using master_lewan_message  = master_lewan_protocol::message_t;

template < lewan_message_enum ID, typename... Args >
inline master_lewan_message make_master_lewan_message( lewan_servo_id sid, const Args&... args )
{
        master_lewan_message res = master_lewan_protocol::create(
            lewan_header_start_v,
            lewan_header_start_v,
            sid,
            master_lewan_group::create< ID >( args... ),
            lewan_checksum{} );

        res.back() = make_lewan_checksum( res );

        return res;
}

}  // namespace schpin
