// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "fw/base.h"
#include "fw/lewan/config.h"
#include "fw/lewan/context.h"
#include "fw/lewan/conv.h"
#include "protocol.h"

#include <emlabcpp/algorithm.h>
#include <emlabcpp/pid.h>
#include <iostream>
#include <schpin_lib/algorithm.h>
#include <schpin_lib/logging.h>
#include <variant>

#pragma once

namespace schpin
{

struct lewan_goto_event
{
        lewan_angle g;
        btime       t;
};

struct lewan_spin_event
{
        lewan_angular_velocity av;
        btime                  t;
};

struct lewan_config_event
{
        lewan_config_enum addr;
        config_buffer     buffr;
};

using lewan_servo_events_variant =
    std::variant< lewan_goto_event, lewan_spin_event, lewan_config_event >;

class lewan_servo
{
        struct activate_task
        {
                std::size_t i = 0;
                cmd_roulette<
                    LEWAN_SERVO_ANGLE_OFFSET_READ,
                    LEWAN_SERVO_ANGLE_LIMIT_READ,
                    LEWAN_SERVO_VIN_LIMIT_READ,
                    LEWAN_SERVO_TEMP_MAX_LIMIT_READ,
                    LEWAN_SERVO_OR_MOTOR_MODE_READ,
                    LEWAN_SERVO_LED_CTRL_READ,
                    LEWAN_SERVO_POS_READ >
                    startup_roulette;
        };

        struct goto_task
        {
                lewan_full_angle start;
                lewan_angle      goal;
                clk_time         start_time;
                btime            time;
        };

        struct spin_task
        {
                clk_time clock_end;
        };

        struct idle_task
        {
        };

        using task_variant = std::variant< activate_task, goto_task, spin_task, idle_task >;

        lewan_config_map                      config_map_;
        lewan_context                         context_;
        lewan_converter                       converter_;
        task_variant                          task_ = { activate_task{} };
        lewan_servo_id                        id_;
        lewan_full_angle                      last_report_;
        message_catcher< lewan_master_group > lewan_catcher_;
        std::size_t                           poll_i_         = 0;
        s_ctl_id                              msg_id_counter_ = 0;

        bool                new_pos_          = false;
        bool                vel_pid_enabled_  = false;
        float               desired_velocity_ = 0.f;
        em::pid< clk_time > vel_pid_;

        bool                pos_pid_enabled_ = false;
        float               desired_pos_     = 0.f;
        em::pid< clk_time > pos_pid_;

public:
        using receive_type   = lewan_master_variant;
        using events_variant = lewan_servo_events_variant;

        lewan_servo( lewan_servo_id id, clk_time now )
          : id_( id )
          , vel_pid_( now )
          , pos_pid_( now )
        {
                configure_pid();
        }

        servo_id get_id() const
        {
                return *id_;
        }

        bool has_task() const
        {
                return !std::holds_alternative< idle_task >( task_ );
        }

        bool is_activating() const
        {
                return std::holds_alternative< activate_task >( task_ );
        }

        const lewan_servo_state& get_state() const
        {
                return *context_;
        }

        template < typename Visitor >
        void stop( Visitor&& vis )
        {
                // TODO: DO NOT DO THIS IF ACTIVATING?
                task_ = idle_task{};
                transmit_servo< LEWAN_SERVO_MOVE_STOP, reply::NO >( vis );
        }

        std::optional< servo_master_message >
        on_servo_receive( clk_time clock_time, s_ctl_id, const lewan_master_variant& var )
        {
                lewan_catcher_.check( var );
                context_.on_servo_receive( clock_time, var );

                using pos_tuple = std::tuple< cmd_tag< LEWAN_SERVO_POS_READ >, lewan_full_angle >;
                if ( !std::holds_alternative< pos_tuple >( var ) ) {
                        return {};
                }

                new_pos_ = true;

                lewan_angle change = config_map_.get_val< LEWAN_ANGLE_CHANGE_REPORT >();
                if ( change == lewan_angle::get< 0 >() ) {
                        return {};
                }

                if ( angle_dist( last_report_, context_->pos ) < *change ) {
                        return {};
                }
                last_report_ = context_->pos;
                return servo_master_group::create< POSITION >(
                    converter_.se_to_in( context_->pos ) );
        }

        template < typename Visitor >
        void tick( clk_time clock_time, Visitor&& vis )
        {
                // TODO:
                //  STALL DETECTION
                if ( new_pos_ ) {
                        new_pos_ = false;
                        if ( pos_pid_enabled_ ) {
                                float actual_pose = static_cast< float >( *context_->pos );
                                pos_pid_.update( clock_time, actual_pose, desired_pos_ );
                                desired_velocity_ = pos_pid_.get_output();
                                SCHPIN_INFO_LOG(
                                    "servo",
                                    "pos PID: " << actual_pose << "," << desired_pos_ << ","
                                                << desired_velocity_ );
                                servo_reply_ctl< TELEM >(
                                    vis,
                                    bangle::from_float( actual_pose ),
                                    bangle::from_float( desired_pos_ ),
                                    bangular_velocity::from_float( desired_velocity_ ) );
                        }
                        if ( vel_pid_enabled_ ) {
                                float vel = static_cast< float >( *context_->velocity );
                                vel_pid_.update( clock_time, vel, desired_velocity_ );
                                SCHPIN_INFO_LOG(
                                    "servo",
                                    "vel PID: " << vel << "\t" << desired_velocity_ << "\t"
                                                << vel_pid_.get_output() );
                                transmit_servo< LEWAN_SERVO_OR_MOTOR_MODE_WRITE, reply::NO >(
                                    vis,
                                    lewan_motor_mode,
                                    uint8_t{ 0 },
                                    *lewan_angular_velocity::make( vel_pid_.get_output() ) );
                        }
                }
                task_ << match(
                    [&]( idle_task ) {},
                    [&]( const activate_task& s ) {
                            activate_tick( s, vis );
                    },
                    [&]( goto_task t ) {
                            // stuff may overflow badyl here
                            float goal_f  = static_cast< float >( *t.goal );
                            float start_f = static_cast< float >( *t.start );
                            desired_pos_  = em::map_range(
                                static_cast< float >( em::min(
                                    *clock_time - *t.start_time,
                                    static_cast< uint32_t >( *t.time ) ) ),
                                0.f,
                                static_cast< float >( *t.time ),
                                start_f,
                                goal_f );
                            task_ = t;
                            if ( *t.start_time + static_cast< uint32_t >( *t.time ) >
                                 *clock_time ) {
                                    return;
                            }
                            // random number - meh TODO
                            if ( angle_dist( context_->pos, t.goal ) < 15 ) {
                                    task_ = idle_task{};
                            }
                    },
                    [&]( const spin_task& t ) {
                            if ( t.clock_end > clock_time ) {
                                    return;
                            }
                            desired_velocity_ = 0.f;
                            task_             = idle_task{};
                    } );
        }

        template < typename Visitor >
        void on_event(
            ctl_id                            cid,
            clk_time                          clock_time,
            const lewan_servo_events_variant& var,
            Visitor&&                         vis )
        {
                // TODO: the states are based on velocity controll in each instance -> to simplify
                // implementation/problems switch to velocity pid control in all of them
                //  - this simplifies setup, as it removes a lot of PID problems
                var << match(
                    [&]( lewan_goto_event e ) {
                            vel_pid_enabled_ = true;
                            vel_pid_.set_time( clock_time );

                            pos_pid_enabled_ = true;
                            pos_pid_.set_time( clock_time );

                            desired_pos_ = *context_->pos;
                            task_        = goto_task{
                                context_->pos,
                                e.g,
                                clock_time,
                                e.t,
                            };
                    },
                    [&]( lewan_spin_event e ) {
                            desired_velocity_ = static_cast< float >( *e.av );

                            vel_pid_enabled_ = true;
                            vel_pid_.set_time( clock_time );

                            pos_pid_enabled_ = false;
                            pos_pid_.reset();

                            task_ = spin_task{ clock_time + clk_time{ e.t } };
                    },
                    [&]( lewan_config_event e ) {  //
                            auto opt_error = config_map_.set_val( e.addr, e.buffr );
                            if ( opt_error ) {
                                    servo_reply_ctl_error< CONFIG_E >( cid, vis );
                            }
                            configure_pid();
                            converter_ = lewan_converter{
                                config_map_.get_val< LEWAN_INPUT_RANGE_LOW >(),
                                config_map_.get_val< LEWAN_INPUT_RANGE_HIGH >(),
                                config_map_.get_val< LEWAN_SERVO_RANGE_LOW >(),
                                config_map_.get_val< LEWAN_SERVO_RANGE_HIGH >() };
                    } );
        }

        template < typename Visitor >
        void poll( clk_time, std::size_t i_offset, Visitor&& vis )
        {
                poll_i_       = ( poll_i_ + 1 ) % 13;
                std::size_t i = ( i_offset + poll_i_ ) % 13;
                switch ( i ) {
                        case 4:
                                transmit_servo< LEWAN_SERVO_VIN_READ, reply::YES >( vis );
                                break;
                        case 8:
                                transmit_servo< LEWAN_SERVO_TEMP_READ, reply::YES >( vis );
                                break;
                        default:
                                transmit_servo< LEWAN_SERVO_POS_READ, reply::YES >( vis );
                                break;
                }
        }

        em::either< lewan_servo_events_variant, servo_master_message >
        on_servo_packet( ctl_id id, master_servo_command_variant subcmd )
        {
                using return_type = em::either< lewan_servo_events_variant, servo_master_message >;

                return subcmd << tuple_match(
                           [&]( cmd_tag< GOTO >, bangle a, btime t ) -> return_type {
                                   return handle_in_to_se( converter_, id, GOTO, a )
                                       .convert_left(
                                           [&]( lewan_angle an ) -> lewan_servo_events_variant {
                                                   SCHPIN_INFO_LOG(
                                                       "servo", "LEWAN GOTO " << an << " , " << t );
                                                   return lewan_goto_event{ an, t };
                                           } );
                           },
                           [&]( cmd_tag< SPIN >, bangular_velocity av, btime t ) -> return_type {
                                   return handle_in_to_se( converter_, id, SPIN, av )
                                       .convert_left(
                                           [&]( lewan_angular_velocity anv )
                                               -> lewan_servo_events_variant {
                                                   SCHPIN_INFO_LOG(
                                                       "servo",
                                                       "LEWAN SPIN " << anv << " , " << t );
                                                   return lewan_spin_event{ anv, t };
                                           } );
                           },
                           [&]( cmd_tag< POSITION > ) -> return_type {
                                   return make_servo_message< POSITION >(
                                       converter_.se_to_in( context_->pos ) );
                           },
                           [&]( cmd_tag< TEMP > ) -> return_type {
                                   return make_servo_message< TEMP >(
                                       converter_.se_to_in( context_->temperature ) );
                           },
                           [&]( cmd_tag< VOLTAGE > ) -> return_type {
                                   return make_servo_message< VOLTAGE >(
                                       converter_.se_to_in( context_->voltage ) );
                           },
                           [&]( cmd_tag< VELOCITY > ) -> return_type {
                                   return make_servo_message< VELOCITY >(
                                       converter_.se_to_in( context_->velocity ) );
                           },
                           [&]( cmd_tag< TASK > ) -> return_type {
                                   servo_task_enum e = task_ << match(
                                                           [&]( idle_task ) {
                                                                   return TASK_IDLE;
                                                           },
                                                           [&]( activate_task ) {
                                                                   return TASK_ACTIVATE;
                                                           },
                                                           [&]( goto_task ) {
                                                                   return TASK_GOTO;
                                                           },
                                                           [&]( spin_task ) {
                                                                   return TASK_SPIN;
                                                           } );

                                   return make_servo_message< TASK >( e );
                           },
                           [&]( cmd_tag< CONFIG_GET >, config_addr addr ) -> return_type {
                                   return make_servo_message< CONFIG_GET >(
                                       addr,
                                       config_map_.get_buffered_val< config_buffer >(
                                           static_cast< lewan_config_enum >( addr ) ) );
                           },
                           [&]( cmd_tag< CONFIG_SET >,
                                config_addr   addr,
                                config_buffer buffr ) -> return_type {
                                   return lewan_servo_events_variant{ lewan_config_event{
                                       static_cast< lewan_config_enum >( addr ), buffr } };
                           } );
        }

        void dump_info()
        {
                std::cout << config_map_;
        }

private:
        template < typename Visitor >
        void activate_tick( activate_task s, Visitor&& v )
        {
                switch ( s.i ) {
                        case 0:
                                s.startup_roulette.tick( [&]( auto tag ) {  //
                                        transmit_servo< decltype( tag )::value, reply::YES >( v );
                                } );
                                if ( s.startup_roulette.get_index() != 0 ) {
                                        task_ = s;
                                        return;
                                }
                                break;
                        case 1:
                                lewan_catcher_.set_trap( LEWAN_SERVO_POS_READ );
                                transmit_servo< LEWAN_SERVO_POS_READ, reply::YES >( v );
                                break;
                        case 2:
                                if ( !lewan_catcher_.catched() ) {
                                        return;
                                }
                                break;
                        default:
                                v( activate_tag{}, *id_ );
                                task_ = idle_task{};
                                return;
                }

                s.i += 1;
                task_ = s;
        }

        template < messages_enum ID, typename Visitor, typename... Args >
        void servo_reply_ctl( Visitor&& v, Args&&... args )
        {
                v( control_tag{}, *id_, servo_master_group::create< ID >( args... ) );
        }

        template < error_protocol_enum ID, typename Visitor, typename... Args >
        void servo_reply_ctl_error( ctl_id cid, Visitor&& v, Args&&... args )
        {
                servo_reply_ctl< ERROR >( v, cid, servo_error_group::create< ID >( args... ) );
        }

        template < lewan_message_enum ID, reply Rpl, typename Visitor, typename... Args >
        void transmit_servo( Visitor&& v, Args... args )
        {
                msg_id_counter_++;
                v( servo_tag{},
                   msg_id_counter_,
                   Rpl,
                   make_master_lewan_message< ID >( id_, args... ) );
        }

        void configure_pid()
        {
                vel_pid_.set_config( em::pid< clk_time >::config{
                    config_map_.get_val< LEWAN_VEL_P_GAIN >().as_float(),
                    config_map_.get_val< LEWAN_VEL_I_GAIN >().as_float(),
                    config_map_.get_val< LEWAN_VEL_D_GAIN >().as_float(),
                    -1000.f,
                    1000.f } );

                pos_pid_.set_config( em::pid< clk_time >::config{
                    config_map_.get_val< LEWAN_GOTO_P_GAIN >().as_float(),
                    config_map_.get_val< LEWAN_GOTO_I_GAIN >().as_float(),
                    config_map_.get_val< LEWAN_GOTO_D_GAIN >().as_float(),
                    -1000.f,
                    1000.f } );
        }
};

}  // namespace schpin
