// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "fw/base.h"
#include "fw/servo/group.h"
#include "shared/protocol.h"

#include <array>
#include <emlabcpp/static_circular_buffer.h>

#pragma once

namespace schpin
{

struct wait_event
{
        servo_flags flags;
};

struct activate_event
{
        servo_addr addr;
};

struct deactivate_event
{
        servo_id sid;
};

struct config_event
{
        config_addr   addr;
        config_buffer buffr;
};

template < unsigned ServoCount >
class fw_reactor
{
        static_assert( ServoCount <= servo_flags().size(),
                       "number of servos is limited based on servo_flags type size" );
        using event_variant =
            std::variant< wait_event, servo_event, activate_event, deactivate_event, config_event >;

        em::static_circular_buffer< std::pair< ctl_id, node_fw_group_variant >, 256 >
                                                                              ictl_buffer_{};
        em::static_circular_buffer< std::pair< ctl_id, event_variant >, 128 > event_buffer_{};
        servo_group< ServoCount >                                             servo_group_{};
        fw_config_map                                                         config_map_;

        clk_time last_input_bufer_telem_{};
        clk_time last_event_bufer_telem_{};

      public:
        template < typename Visitor >
        void on_servo_receive( clk_time                    clock_time,
                               s_ctl_id                    s_id,
                               const lewan_master_message& m,
                               Visitor&&                   v )
        {
                servo_group_.on_servo_receive( clock_time, s_id, m, v );
        }

        template < typename Message, typename Visitor >
        void on_servo_not_replying( const Message& m, Visitor&& v )
        {
                servo_group_.on_servo_not_replying( m, v );
        }

        template < typename Visitor >
        void on_control_receive( ctl_id id, const node_fw_group_variant& m, Visitor&& v )
        {
                if ( ictl_buffer_.full() ) {
                        reply_ctl_error< INPUT_CONTROL_BUFFER_FULL_E >( id, v );
                }
                ictl_buffer_.push_back( std::make_pair( id, m ) );
        }

        template < typename Visitor >
        void tick( clk_time clock_time, Visitor&& v )
        {
                for ( std::size_t i = 0; i < 2; ++i ) {
                        if ( !ictl_buffer_.empty() ) {
                                auto [id, var] = ictl_buffer_.pop_front();
                                on_packet( id, var, v );
                        }
                }

                handle_event( clock_time, v );

                if ( last_input_bufer_telem_ +
                         config_map_.get_val< INPUT_BUFFER_TELEM_PERIOD_MS >() <=
                     clock_time ) {
                        reply_ctl< INPUT_BUFFER_SIZE >(
                            v, static_cast< uint16_t >( ictl_buffer_.size() ) );
                        last_input_bufer_telem_ = clock_time;
                }
                if ( last_input_bufer_telem_ +
                         config_map_.get_val< EVENT_BUFFER_TELEM_PERIOD_MS >() <=
                     clock_time ) {
                        reply_ctl< EVENT_BUFFER_SIZE >(
                            v, static_cast< uint16_t >( event_buffer_.size() ) );
                        last_event_bufer_telem_ = clock_time;
                }

                servo_group_.tick( clock_time, v );
        }

        template < typename Visitor >
        void handle_event( clk_time clock_time, Visitor&& v )
        {
                servo_flags blocking_flags;  // flags where b[i] = 1 if 'i' servo is blocked by some
                                             // event in the queue
                blocking_flags.reset();
                auto iter = em::find_if(
                    event_buffer_, em::uncurry | [&]( ctl_id, const event_variant& ev ) {
                            return ev << match(
                                       [&]( wait_event e ) {
                                               if ( ( e.flags & blocking_flags ).any() ) {
                                                       // previous wait_event is blocking some servo
                                                       // of this event, so this has to be skipped.
                                                       blocking_flags |= e.flags;
                                                       return false;
                                               }
                                               // now we know that none of the e.flags servos is
                                               // blocked by previous wait events
                                               if ( servo_group_.any_has_task( e.flags ) ) {
                                                       // at least one of the servos is busy
                                                       blocking_flags |= e.flags;
                                                       return false;
                                               }
                                               // all servos are finished and there is no previous
                                               // blocking event, this is it!
                                               return true;
                                       },
                                       [&]( servo_event e ) {
                                               return !blocking_flags[e.id];
                                       },
                                       [&]( activate_event ) {
                                               return true;
                                       },
                                       [&]( deactivate_event ) {
                                               return true;
                                       },
                                       [&]( config_event ) {
                                               return true;
                                       } );
                    } );
                if ( iter == event_buffer_.end() ) {
                        return;
                }
                ctl_id        id    = iter->first;
                event_variant event = iter->second;
                SCHPIN_INFO_LOG( "reactor",
                                 "finishing event: " << id << " used: " << event_buffer_.size()
                                                     << "/" << event_buffer_.max_size() );
                // move selected event to the front of reversed(buffer) -> end of the buffer
                while ( iter != event_buffer_.begin() ) {
                        using namespace std;
                        swap( *std::prev( iter ), *iter );
                        iter--;
                }
                event_buffer_.pop_front();

                event << match(
                    [&]( wait_event e ) {  //
                            SCHPIN_INFO_LOG( "reactor", "flags of finished: " << e.flags );
                            reply_ctl< FINISHED >( v, id );
                    },
                    [&]( servo_event e ) {  //
                            servo_group_.on_event( id, clock_time, e, v );
                    },
                    [&]( activate_event e ) {
                            servo_group_.activate( clock_time, id, e.addr, v );
                    },
                    [&]( deactivate_event e ) {
                            servo_group_.deactivate( id, e.sid, v );
                    },
                    [&]( config_event e ) {
                            // TODO: error handling
                            // TODO: look at all places where register_map is used and revamp API
                            auto opt_error = config_map_.set_val(
                                static_cast< fw_config_enum >( e.addr ), e.buffr );
                            if ( opt_error ) {
                                    reply_ctl_error< CONFIG_E >( id, v );
                            }
                    } );
        }

        template < typename Visitor >
        void on_packet( ctl_id id, node_fw_group_variant var, Visitor&& v )
        {
                var << tuple_match(
                    [&]( cmd_tag< PING > ) {  //
                            reply_ctl< PONG >( v );
                    },
                    [&]( cmd_tag< STOP > ) {  //
                            servo_group_.stop( v );
                    },
                    [&]( cmd_tag< SYNC >, servo_flags flag ) {  //
                            add_event( id, wait_event{ flag }, v );
                    },
                    [&]( cmd_tag< SERVO >, servo_id sid, master_servo_command_variant subvar ) {
                            servo_group_.on_servo_packet( id, sid, subvar )
                                .match(
                                    [&]( servo_event e ) {  //
                                            add_event( id, e, v );
                                    },
                                    [&]( fw_node_group_message msg ) {
                                            v( control_tag{}, msg );
                                    } );
                    },
                    [&]( cmd_tag< ACTIVATE >, servo_addr sid ) {
                            add_event( id, activate_event{ sid }, v );
                    },
                    [&]( cmd_tag< DEACTIVATE >, servo_id sid ) {
                            add_event( id, deactivate_event{ sid }, v );
                    },
                    [&]( cmd_tag< CONFIG_SET >, config_addr addr, config_buffer bffr ) {
                            add_event( id, config_event{ addr, bffr }, v );
                    },
                    [&]( cmd_tag< CONFIG_GET >, config_addr addr ) {
                            reply_ctl< CONFIG_GET >( v,
                                                     addr,
                                                     config_map_.get_buffered_val< config_buffer >(
                                                         static_cast< fw_config_enum >( addr ) ) );
                    } );
        }

      private:
        template < typename Visitor >
        void add_event( ctl_id id, event_variant e, Visitor&& v )
        {
                if ( event_buffer_.full() ) {
                        reply_ctl_error< INPUT_EVENT_BUFFER_FULL_E >( id, v );
                }
                SCHPIN_INFO_LOG( "info",
                                 "adding event: " << id << " used: " << event_buffer_.size() << "/"
                                                  << event_buffer_.max_size() );
                event_buffer_.push_back( std::make_pair( id, e ) );
        }
};

}  // namespace schpin
