// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include <errno.h>
#include <fcntl.h>
#include <filesystem>
#include <schpin_lib/error.h>
#include <termios.h>
#include <unistd.h>

#pragma once

namespace schpin
{

using serial_error = error< struct serial_error_tag >;

class serial_connection
{
        int fd_;

        serial_connection( int fd )
          : fd_( fd )
        {
        }

      public:
        static em::either< serial_connection, serial_error >
        make( const std::filesystem::path& filename )
        {
                int fd = open( filename.c_str(), O_RDWR );

                if ( fd < 0 ) {
                        return { E( serial_error ) << "Error " << errno << " when opening file "
                                                   << filename << " :" << strerror( errno ) };
                }

                struct termios tty = {};

                if ( tcgetattr( fd, &tty ) != 0 ) {
                        return { E( serial_error )
                                 << "Error " << errno << " when getting attributes of console "
                                 << filename << " :" << strerror( errno ) };
                }

                auto set = []( auto& flags, unsigned flag ) {
                        flags |= flag;
                };
                auto clr = []( auto& flags, unsigned flag ) {
                        flags &= ~flag;
                };

                clr( tty.c_cflag, PARENB );
                clr( tty.c_cflag, CSTOPB );
                set( tty.c_cflag, CS8 );
                clr( tty.c_cflag, CRTSCTS );
                set( tty.c_cflag, CREAD | CLOCAL );

                clr( tty.c_lflag, ICANON );
                clr( tty.c_lflag, ECHO );
                clr( tty.c_lflag, ECHOE );
                clr( tty.c_lflag, ECHONL );
                clr( tty.c_lflag, ISIG );
                clr( tty.c_iflag, IXON | IXOFF | IXANY );
                clr( tty.c_iflag, IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR | IGNCR | ICRNL );

                clr( tty.c_oflag, OPOST );
                clr( tty.c_oflag, ONLCR );

                tty.c_cc[VTIME] = 2;  // 2 deciseconds
                tty.c_cc[VMIN]  = 0;

                cfsetispeed( &tty, B115200 );
                cfsetospeed( &tty, B115200 );

                if ( tcsetattr( fd, TCSANOW, &tty ) != 0 ) {
                        return { E( serial_error )
                                 << "Error " << errno << " when settings attributes of console "
                                 << filename << " :" << strerror( errno ) };
                }

                return serial_connection{ fd };
        }

        template < typename Message >
        void write( const Message& m )
        {
                ::write( fd_, m.begin(), m.size() );
        }

        template < std::size_t N >
        em::either< em::static_vector< uint8_t, N >, serial_error > read()
        {
                std::array< uint8_t, N > buff;
                ssize_t                  n = ::read( fd_, buff.begin(), buff.size() );

                if ( n < 0 ) {
                        return { E( serial_error )
                                 << "Error " << errno
                                 << " when reading the console:" << strerror( errno ) };
                }

                em::static_vector< uint8_t, N > res;

                std::copy_n( buff.begin(), n, std::back_inserter( res ) );

                return { res };
        }

        ssize_t read( uint8_t* iter, std::size_t n )
        {
                return ::read( fd_, iter, n );
        }
};

}  // namespace schpin
