// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "fw/base.h"
#include "fw/dynmx/config.h"
#include "fw/lewan/config.h"

#include <emlabcpp/physical_quantity.h>
#include <schpin_lib_exp/config.h>

#pragma once

namespace schpin
{

template <>
struct ros_param_serializer< servo_type > : ros_param_serializer_enum< servo_type >
{
};

static auto from_angle_param = ros_opt_param< bangle >(
    "from_angle",
    "angle at which movement starts",
    bangle::from_float( 0.f ) );

static auto to_angle_param = ros_opt_param< bangle >(
    "to_angle",
    "angle at which movement ends",
    bangle::from_float( 4.18879f ) );

template < typename T >
inline auto make_from_servo_param( T def )
{
        return ros_opt_param< T >( "from_servo", "servo angle at which movement starts", def );
}

template < typename T >
inline auto make_to_servo_param( T def )
{
        return ros_opt_param< T >( "to_servo", "servo angle at which movement ends", def );
}

template < typename T >
inline auto make_angle_change_param( T def )
{
        return ros_opt_param< T >(
            "angle_change",
            "change of angle that results in automatic position report, use servo units",
            def );
}

struct node_lewan_servo_config
{
        lewan_config_map cmap;

        template < typename... Ts >
        node_lewan_servo_config( Ts... items )
          : cmap( items... )
        {
        }

        static auto ros_param_get()
        {
                return std::make_tuple(
                    from_angle_param,
                    to_angle_param,
                    make_from_servo_param< lewan_angle >( lewan_angle::get< 0 >() ),
                    make_to_servo_param< lewan_angle >( lewan_angle::get< 0 >() ),
                    make_angle_change_param< lewan_angle >( lewan_angle::get< 0 >() ),
                    ros_opt_param< lewan_angle >( "stall_range", "???", lewan_angle::get< 5 >() ),
                    ros_param< pid_gain >( "goto_p", "P gain in goto PID regulator" ),
                    ros_param< pid_gain >( "goto_i", "I gain in goto PID regulator" ),
                    ros_param< pid_gain >( "goto_d", "D gain in goto PID regulator" ),
                    ros_param< pid_gain >( "vel_p", "P gain in vel PID regulator" ),
                    ros_param< pid_gain >( "vel_i", "I gain in vel PID regulator" ),
                    ros_param< pid_gain >( "vel_d", "D gain in vel PID regulator" ) );
        }
};

struct node_dynmx_servo_config
{
};

using node_servo_config_var = std::variant< lewan_config_map, dynmx_config_map >;

struct node_servo_config
{
        servo_id              id;
        servo_type            stype;
        std::string           joint_name;
        node_servo_config_var cvar;
};

template <>
struct ros_param_serializer< node_servo_config >
{

        static auto get_var( rclcpp::Node& node, servo_type stype, const std::string& prefix )
        {
                switch ( stype ) {
                        case LEWAN:
                        case LEWAN_TEST:
                                return ros_param_serializer< node_lewan_servo_config >::load(
                                           node, prefix )
                                    .convert_left(
                                        [&](
                                            node_lewan_servo_config cfg ) -> node_servo_config_var {
                                                return cfg.cmap;
                                        } );
                        case DYNMX:
                                return em::either< node_servo_config_var, config_error >{
                                    dynmx_config_map{} };
                }
                return em::either< node_servo_config_var, config_error >{
                    E( config_error ) << "Unknown servo type for config laod" };
        }
        static em::either< node_servo_config, config_error >
        load( rclcpp::Node& node, const std::string& prefix )
        {
                return em::assemble_left_collect_right(
                           ros_param< servo_id >( "id", "id of the servo on its bus" )
                               .load( node, prefix ),
                           ros_param< servo_type >( "type", "type of the servomotor" )
                               .load( node, prefix ),
                           ros_param< std::string >(
                               "joint_name", "name of the joint that is used for this servo" )
                               .load( node, prefix ) )
                    .convert_right( [&]( const auto& errs ) {
                            return E( config_error ).group_attach( errs )
                                << "Failed to get servo config header";
                    } )
                    .bind_left(
                        em::uncurry | [&]( servo_id sid, servo_type stype, std::string jname ) {
                                return get_var( node, stype, prefix )
                                    .convert_left( [&]( node_servo_config_var cvar ) {
                                            return node_servo_config{ sid, stype, jname, cvar };
                                    } );
                        } );
        }
};

struct node_hardware_config
{
        std::map< std::string, node_servo_config > servos;
        std::string                                diag_topic;
        float                                      diag_freq;
        std::string                                joint_state_topic;
        std::string                                robot_description;
        std::string                                robot_semantics;

        static auto ros_param_get()
        {
                return std::make_tuple(  //
                    ros_param< std::map< std::string, node_servo_config > >(
                        "servos", "definitions of servomotors" ),
                    ros_param< std::string >( "diag_topic", "Diagnostics topic" ),
                    ros_param< float >( "diag_freq", "Diagnostic frequency" ),
                    ros_param< std::string >(
                        "joint_state", "Topic at which send updates of the joint positions." ),
                    ros_param< std::string >( "urdf", "URDF description" ),
                    ros_param< std::string >( "srdf", "SRDF descriptoon" ) );
        }
};

}  // namespace schpin
