// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "fw/lewan/config.h"
#include "node/base.h"
#include "node/config.h"
#include "shared/protocol.h"
#include <magic_enum.hpp>

#pragma once

namespace schpin
{

class servo
{
        std::string           name_;
        std::string           joint_name_;
        servo_id              id_;
        servo_type            stype_;
        node_servo_config_var config_var_;
        bool      config_changed_ = true;  // default true is important - config set after init!
        em::angle state_;

public:
        servo( std::string name, const node_servo_config& c )
          : name_( std::move( name ) )
          , joint_name_( c.joint_name )
          , id_( c.id )
          , stype_( c.stype )
          , config_var_( c.cvar )
        {
        }

        const node_servo_config_var& get_config_var() const
        {
                return config_var_;
        }

        const std::string& get_name() const
        {
                return name_;
        }

        const std::string& get_joint_name() const
        {
                return joint_name_;
        }

        const servo_id& get_id() const
        {
                return id_;
        }

        void set_config( const nlohmann::json& j )
        {
                std::visit(
                    [&]( auto& conf ) {
                            from_json( j, conf );
                            SCHPIN_INFO_LOG(
                                "wololo", "setting config to: " << nlohmann::json{ conf } );
                    },
                    config_var_ );
                config_changed_ = true;
        }

        node_fw_group_message make_pos_message() const
        {
                return make_message< POSITION >();
        }

        node_fw_group_message make_goto_message( bangle a, btime t ) const
        {
                return make_message< GOTO >( a, t );
        }

        node_fw_group_message make_spin_message( bangular_velocity av, btime t ) const
        {
                return make_message< SPIN >( av, t );
        }

        node_fw_group_message make_activate_message() const
        {
                // TODO: this actually may be an error
                return node_fw_group::create< ACTIVATE >( servo_addr{ LEWAN, id_ } );
        }

        template < typename Visitor >
        void on_feedback( servo_master_command_variant v, Visitor&& vis )
        {
                // this is not a good approach, the "boot" of servos is not well defined in APP,
                // there should be proper process for it -> servo is actiavted, than config is
                // loaded
                if ( config_changed_ ) {
                        config_changed_ = false;
                        upload_config( vis );
                }

                v << tuple_match(  //
                    [&]( cmd_tag< POSITION >, bangle a ) {
                            state_ = em::angle{ a.as_float() };
                            vis( *this, state_ );
                    },
                    [&]( auto... ) {} );
        }

private:
        template < typename Visitor >
        void upload_config( Visitor&& vis )
        {
                std::visit(
                    [&]( const auto& cmap ) {
                            for_each_reg( cmap, [&]( auto tok ) {
                                    static constexpr auto reg = decltype( tok )::reg;
                                    SCHPIN_INFO_LOG(
                                        "servo",
                                        "Setting config val: " << reg << " "
                                                               << cmap.template get_val< reg >() );
                                    vis( make_message< CONFIG_SET >(
                                        reg,
                                        cmap.template get_buffered_val< config_buffer >( reg ) ) );
                            } );
                    },
                    config_var_ );
        }
        template < messages_enum CMD, typename... Args >
        node_fw_group_message make_message( Args... args ) const
        {
                return { node_fw_group::create< SERVO >(
                    id_, master_servo_group::create< CMD >( args... ) ) };
        }
};

inline std::ostream& operator<<( std::ostream& os, const servo& s )
{
        os << s.get_name() << " - " << s.get_joint_name() << " - " << s.get_id() << "\n";
        std::visit(
            [&]( const auto& conf ) {
                    os << conf;
            },
            s.get_config_var() );
        return os;
}

class servo_group
{
        std::vector< servo > servos_;

public:
        servo_group( const std::map< std::string, node_servo_config >& configs )
          : servos_(
                map_f_to_v( configs, [&]( const std::pair< std::string, node_servo_config >& cp ) {
                        return servo{ cp.first, cp.second };
                } ) )
        {
        }

        const std::vector< servo >& get_servos() const
        {
                return servos_;
        }

        auto find_servo( servo_id id )
        {
                return em::find_if( servos_, [&]( const servo& s ) {  //
                        return s.get_id() == id;
                } );
        }

        void set_servo_config( servo_id id, const nlohmann::json& j )
        {
                auto iter = find_servo( id );
                if ( iter == servos_.end() ) {
                        SCHPIN_ERROR_LOG(
                            "servos", "Got servo config command for non-existing servo: " << id );
                        return;
                }
                iter->set_config( j );
        }

        template < typename Visitor >
        void on_feedback( servo_id id, servo_master_command_variant v, Visitor&& vis )
        {
                auto iter = find_servo( id );
                if ( iter == servos_.end() ) {
                        SCHPIN_ERROR_LOG(
                            "servos", "Got servo command for non-existing servo: " << id );
                        return;
                }

                iter->on_feedback( v, vis );
        }
};

inline std::ostream& operator<<( std::ostream& os, const servo_group& g )
{
        os << "servo group: " << std::endl;
        for ( const servo& s : g.get_servos() ) {
                os << s << std::endl;
        }
        return os;
}

}  // namespace schpin
