// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "fw/lewan/protocol.h"

#include <gtest/gtest.h>
#include <schpin_lib/logging.h>
#include <schpin_lib/view.h>

using namespace schpin;

TEST( lewan_protocol, master_lewan )
{

        master_lewan_message msg =
            make_master_lewan_message< LEWAN_SERVO_POS_READ >( lewan_servo_id::get< 8 >() );
        master_lewan_message expected{
            std::array< uint8_t, 6 >{ 0x55, 0x55, 8, 3, LEWAN_SERVO_POS_READ, 216 } };

        EXPECT_EQ( msg, expected );

        master_lewan_protocol::extract( msg ).match(
            em::uncurry |
                [&]( lewan_header_start,
                     lewan_header_start,
                     lewan_servo_id id,
                     auto           parsed_var,
                     lewan_checksum ) {
                        EXPECT_EQ( id, lewan_servo_id::get< 8 >() );

                        bool cmd_tuple =
                            std::holds_alternative< std::tuple< cmd_tag< LEWAN_SERVO_POS_READ > > >(
                                parsed_var );
                        EXPECT_TRUE( cmd_tuple );
                },
            [&]( auto e ) {
                    FAIL() << e;
            } );
}

TEST( lewan_protocol, lewan_master )
{
        auto submsg =
            lewan_master_group::create< LEWAN_SERVO_POS_READ >( lewan_angle::get< 500 >() );

        lewan_master_message msg = lewan_master_protocol::create(
            lewan_header_start_v,
            lewan_header_start_v,
            lewan_servo_id::get< 8 >(),
            submsg,
            lewan_checksum{ 225 } );
        lewan_master_message expected{
            std::array< uint8_t, 8 >{ 0x55, 0x55, 8, 5, LEWAN_SERVO_POS_READ, 244, 1, 225 } };

        EXPECT_EQ( msg, expected );

        lewan_master_protocol::extract( msg ).match(
            em::uncurry |
                [&]( lewan_header_start,
                     lewan_header_start,
                     lewan_servo_id id,
                     auto,
                     lewan_checksum ) {
                        EXPECT_EQ( id, lewan_servo_id::get< 8 >() );
                },
            [&]( auto e ) {
                    FAIL() << e;
            } );
}

TEST( lewan_protocol, full_range )
{
        // this is message setn by servo, not valid based on doc.
        lewan_master_message valid{
            std::array< uint8_t, 8 >{ 0x55, 0x55, 3, 5, 28, 103, 255, 117 } };

        lewan_master_protocol::extract( valid ).match(
            [&]( auto ) {},
            [&]( auto e ) {
                    FAIL() << e;
            } );
}

TEST( lewan_protocol, sequencer )
{
        using seqe = lewan_sequencer< lewan_master_message >;

        auto msg = lewan_master_protocol::create(
            lewan_header_start_v,
            lewan_header_start_v,
            lewan_servo_id::get< 1 >(),
            lewan_master_group::create< LEWAN_SERVO_POS_READ >( lewan_angle::get< 500 >() ),
            lewan_checksum{} );

        seqe sequencer{};

        std::size_t fired_n = 0;
        auto        f       = [&]( lewan_master_message submsg ) {  //
                EXPECT_EQ( submsg, msg );
                fired_n += 1;
        };

        // simple load
        sequencer.load( msg, f );
        EXPECT_EQ( fired_n, 1 );

        // load message as multiple parts
        em::view first_part_view{ msg.begin(), msg.begin() + 2 };
        em::view rest_view{ msg.begin() + 2, msg.end() };

        sequencer.load( first_part_view, f );
        EXPECT_EQ( fired_n, 1 );
        sequencer.load( rest_view, f );
        EXPECT_EQ( fired_n, 2 );

        // introduce random bytes at the beginning
        std::array< uint8_t, 4 > garbage{ 0x44, 0x55, 0x66, 0x22 };

        sequencer.load( garbage, f );
        EXPECT_EQ( fired_n, 2 );

        sequencer.load( msg, f );
        EXPECT_EQ( fired_n, 3 );
}

int main( int argc_in, char** argv_in )
{
        testing::InitGoogleTest( &argc_in, argv_in );

        return RUN_ALL_TESTS();
}
