// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "fw/dynmx/serial.h"
#include "fw/serial.h"
#include "fw/servo/holder.h"
#include "node/lua_vm/def.h"
#include "node/lua_vm/extra_ser.h"

#include <rclcpp/rclcpp.hpp>
#include <schpin_lib/lua/repl.h>
#include <schpin_lib/lua/util.h>
#include <schpin_lib/lua/vm.h>
#include <schpin_lib_exp/config.h>

using namespace schpin;

using runner_error = error< struct runner_error_tag >;

// TODO:
//
//    make extension to protocol that allows json -> protocol serialization

// TODO:
//    make sleep function for tests
//    properly handle errors in run_file, now it loops infiniitely

struct msg_event
{
        nlohmann::json j;
};

struct msg_bind
{

        static constexpr std::string_view name = "msg";
        static constexpr std::string_view desc = "creates message out of json";

        inline static const auto arg =
            slua_struct< msg_event >( lua_key_arg< nlohmann::json >{ "json", "data" } );
        inline static const auto res = slua_result< nlohmann::json >{ "Response message" };
};

struct dump_event
{
};

struct dump_bind
{
        static constexpr std::string_view name = "dump";
        static constexpr std::string_view desc = "dump information about servo";

        inline static const auto           arg = slua_struct< dump_event >();
        inline static const no_return_type res{};
};

struct quit_event
{
};

struct quit_bind
{
        static constexpr std::string_view name = "quit";
        static constexpr std::string_view desc = "closes the app";

        inline static const auto           arg = slua_struct< quit_event >();
        inline static const no_return_type res{};
};

struct slua_runner_def
{
        static auto get_def()
        {
                return std::make_tuple(
                    slua_callback< msg_bind >{},
                    slua_callback< dump_bind >{},
                    slua_callback< quit_bind >() );
        }
};

template < auto ID >
struct nlohmann::adl_serializer< cmd_tag< ID > >
{
        static void to_json( nlohmann::json& j, cmd_tag< ID > )
        {
                j = magic_enum::enum_name( ID );
        }
        static void from_json( const nlohmann::json& j, cmd_tag< ID >& id )
        {

                auto opt_val = magic_enum::enum_cast< decltype( ID ) >( j.get< std::string >() );
                if ( !opt_val.has_value() ) {
                        throw nlohmann::json::parse_error::create( -1, 0, "Failed to parse ID" );
                }

                if ( opt_val.value() != ID ) {
                        throw nlohmann::json::parse_error::create( -1, 0, "ID does not match" );
                }

                id = cmd_tag< ID >{};
        }
};

// TODO: scane entire source code for "extra" json serializers and protocol serializers outside of
// their lib
template < std::size_t N >
struct nlohmann::adl_serializer< prot_sizeless_message< N > >
{
        static void to_json( nlohmann::json& j, const prot_sizeless_message< N >& msg )
        {
                for ( uint8_t b : msg ) {
                        j.push_back( b );
                }
        }

        static prot_sizeless_message< N > from_json( const nlohmann::json& j )
        {
                if ( j.size() > N ) {
                        throw nlohmann::json::parse_error::create(
                            -1, 0, "sizeless message in json too big" );
                }
                em::static_vector< uint8_t, N > buffr;
                for ( uint8_t val : j ) {
                        buffr.push_back( val );
                }
                return prot_sizeless_message< N >{ buffr };
        }
};

template < typename >
struct variant_serializer;

template < typename... Ts >
struct variant_serializer< std::variant< Ts... > >
{
        template < typename T >
        struct tag
        {
                using value_type = T;
        };

        static void to_json( nlohmann::json& j, const std::variant< Ts... >& var )
        {
                j = nlohmann::json::array();
                std::visit(
                    [&]( const auto& tpl ) {
                            // TODO: this has to exists because fuckin nlohmann serializes to
                            // std::tuple of size 2 intoobjeck with single key/val instance ;(
                            em::for_each( tpl, [&]( const auto& subitems ) {
                                    j.push_back( subitems );
                            } );
                    },
                    var );
        }

        static std::variant< Ts... > from_json( const nlohmann::json& j )
        {
                std::optional< std::variant< Ts... > > opt_res;

                auto f = [&]( auto tagv ) {
                        using T = typename decltype( tagv )::value_type;
                        try {
                                opt_res = std::variant< Ts... >{ j.get< T >() };
                                return true;
                        }
                        catch ( ... ) {
                                return false;
                        }
                };

                ( f( tag< Ts >{} ) || ... || false );
                if ( !opt_res ) {
                        throw nlohmann::json::parse_error::create(
                            -1, 0, "no variant alternative can be parsed" );
                }
                return *opt_res;
        }
};

template <>
struct nlohmann::adl_serializer< master_servo_command_variant >
  : variant_serializer< master_servo_command_variant >
{
};

template <>
struct nlohmann::adl_serializer< servo_master_command_variant >
  : variant_serializer< servo_master_command_variant >
{
};

template <>
struct nlohmann::adl_serializer< servo_error_variant > : variant_serializer< servo_error_variant >
{
};

struct servo_tester_config
{
        servo_type                             type;
        std::optional< servo_id >              opt_id;
        std::optional< std::filesystem::path > opt_dev;
        std::optional< std::filesystem::path > opt_source;

        static auto ros_param_get()
        {
                return std::make_tuple(  //
                    ros_param< servo_type >( "stype", "Servo type" ),
                    ros_param< std::optional< servo_id > >( "sid", "Servo id" ),
                    ros_param< std::optional< std::filesystem::path > >(
                        "device", "Device to open for connection" ),
                    ros_param< std::optional< std::filesystem::path > >(
                        "script", "Script to execute instead of the interactive mode" ) );
        }
};

class servo_runner
{
        std::optional< dynmx_serial >      dynmx_ser_;
        std::optional< serial_connection > lewan_ser_;
        servo_holder                       servo_;
        slua_vm< slua_runner_def >         vm_;
        slua_repl                          repl_;

        struct visitor
        {
                servo_runner& runner;

                template < typename T >
                void operator()( control_tag, servo_id, T )
                {
                }

                void operator()( activate_tag, servo_id id )
                {
                        SCHPIN_INFO_LOG( "lewan_servo", "Got activation of: " << id );
                }

                void
                operator()( servo_tag, s_ctl_id msg_id, reply r, const master_lewan_message& m )
                {
                        SCHPIN_ASSERT( runner.lewan_ser_ );

                        runner.lewan_ser_->write( m );
                        if ( r == reply::NO ) {
                                return;
                        }
                        runner.receive_serial_msg( msg_id );
                }

                void operator()( servo_tag, s_ctl_id msg_id, reply, const master_dynmx_message& m )
                {
                        SCHPIN_ASSERT( runner.dynmx_ser_ );

                        runner.dynmx_ser_->send( m );

                        dynmx_master_message msg = runner.dynmx_ser_->read_reply();

                        dynmx_master_protocol::extract( msg ).match(
                            em::uncurry |
                                [&]( dynmx_header, dynmx_servo_id, auto tpl, dynmx_crc ) {
                                        const dynmx_response_error& err = std::get< 1 >( tpl );
                                        if ( err != dynmx_no_error ) {
                                                SCHPIN_ERROR_LOG(
                                                    "runner", "Dynmx return error: " << err );
                                        }
                                        const dynmx_buffer& buffer = std::get< 2 >( tpl );

                                        if ( buffer.size() != 0 ) {
                                                runner.servo_.on_servo_receive(
                                                    get_system_time(), msg_id, buffer );
                                        }
                                },
                            [&]( auto err ) {
                                    SCHPIN_ERROR_LOG( "servo_runner", err << ", msg: " << msg );
                            } );
                }
        };

public:
        servo_runner( const servo_tester_config& conf, std::optional< serial_connection > opt_ser )
          : servo_( *servo_holder::make( conf.type, *conf.opt_id, get_system_time() ) )
        {
                SCHPIN_ASSERT( conf.opt_id );
                if ( conf.type == DYNMX ) {
                        SCHPIN_ASSERT( opt_ser );
                        dynmx_ser_.emplace( *opt_ser );
                }
                while ( servo_.is_activating() ) {
                        servo_.tick( get_system_time(), visitor{ *this } );
                }
        }

        void receive_serial_msg( s_ctl_id msg_id )
        {
                static constexpr std::size_t n = lewan_master_message::max_size;

                SCHPIN_ASSERT( lewan_ser_ );

                lewan_ser_->read< n >().match(
                    [&]( const em::static_vector< uint8_t, n >& data ) {
                            lewan_sequencer< lewan_master_message > sequencer;

                            bool load_fired = sequencer.load( data, [&]( lewan_master_message m ) {
                                    parse_message( msg_id, m );
                            } );

                            if ( !load_fired ) {
                                    SCHPIN_ERROR_LOG(
                                        "control",
                                        "servo not replying, got:" << convert_view< int >( data ) );
                            }
                    },
                    [&]( auto err ) {
                            SCHPIN_ERROR_LOG( "control", "Read error: " << err );
                    } );
        }
        void parse_message( s_ctl_id msg_id, const lewan_master_message& m )
        {
                lewan_master_protocol::extract( m ).match(
                    em::uncurry |
                        [&]( lewan_header_start,
                             lewan_header_start,
                             lewan_servo_id,
                             lewan_master_variant msg_var,
                             lewan_checksum ) {
                                auto clk_time = get_system_time();
                                SCHPIN_INFO_LOG( "servo", "clktime: " << clk_time );
                                servo_.on_servo_receive( clk_time, msg_id, msg_var );
                        },
                    [&]( auto err ) {
                            SCHPIN_ERROR_LOG( "lewan_servo", "Lewan prot error: " << err );
                    } );
        }

        void tick()
        {
                servo_.tick( get_system_time(), visitor{ *this } );
                servo_.poll( get_system_time(), 0, visitor{ *this } );
                std::string msg = repl_.prompt();
                vm_.load( msg ).optionally_log();
                vm_.full_exec( [&]( lua_State* lua_state, auto event ) {
                           return handle_event( lua_state, event );
                   } )
                    .optionally_log();
        }

        void run_file( const std::filesystem::path& file )
        {
                auto opt_e = vm_.load( file );
                if ( opt_e ) {
                        opt_e.optionally_log();
                        return;
                }
                bool finished = false;

                while ( !finished ) {
                        servo_.tick( get_system_time(), visitor{ *this } );
                        servo_.poll( get_system_time(), 0, visitor{ *this } );
                        vm_.exec( [&]( lua_State* lua_ptr, auto event ) {
                                return handle_event( lua_ptr, event );
                        } )
                            << match(
                                   [&]( lua_res res ) {
                                           finished = res == lua_res::FINISHED;
                                   },
                                   [&]( slua_error err ) {
                                           finished = true;
                                           SCHPIN_ERROR_LOG( "runner", "Slua error: " << err );
                                   } );
                }
        }

        void handle_event( lua_State*, quit_event )
        {
                std::exit( 0 );
        }

        void handle_event( lua_State*, dump_event )
        {
                servo_.dump_info();
        }

        nlohmann::json handle_event( lua_State* lua_ptr, msg_event e )
        {
                nlohmann::json res;

                // TODO: write proper thing for this in schpin_lib/json and check in rest of the
                // codebase
                try {
                        auto var = e.j.get< master_servo_command_variant >();

                        servo_.on_servo_packet( ctl_id{ 0 }, var )
                            .match(
                                [&]( servo_event se ) {
                                        servo_.on_event(
                                            ctl_id{ 0 },
                                            get_system_time(),
                                            se.sub_event,
                                            visitor{ *this } );
                                },
                                [&]( servo_master_message m ) {
                                        servo_master_group::extract( m ).match(
                                            [&]( auto var ) {
                                                    res = var;
                                            },
                                            [&]( auto err ) {
                                                    SCHPIN_ERROR_LOG(
                                                        "robot",
                                                        "this should not happen error: " << err );
                                            } );
                                } );
                }
                catch ( const nlohmann::json::exception& ex ) {
                        lua_error_stream( lua_ptr, "Json error: ", ex.what() );
                }
                return res;
        }
};

template <>
struct schpin::ros_param_serializer< servo_type > : schpin::ros_param_serializer_enum< servo_type >
{
};

em::either< servo_id, runner_error > find_dynmx_servo( serial_connection ser )
{
        dynmx_serial                    dser( ser );
        auto                            id = dynmx_servo_id::get< 0 >();
        std::optional< dynmx_servo_id > opt_res;

        dser.send( make_master_dynmx_msg< DYNMX_PING >( id ) );

        dynmx_master_message reply = dser.read_reply();

        dynmx_master_protocol::extract( reply ).match(
            em::uncurry |
                [&]( dynmx_header, dynmx_servo_id sid, auto, dynmx_crc ) {
                        opt_res = sid;
                },
            [&]( auto err ) {
                    SCHPIN_ERROR_LOG( "servo_runner", err << ", msg: " << reply );
            } );

        if ( !opt_res ) {
                return E( runner_error ) << "Failed to find id";
        }
        return servo_id{ **opt_res };
}

em::either< servo_id, runner_error > find_servo( servo_type stype, serial_connection ser )
{
        switch ( stype ) {
                case DYNMX:
                        return find_dynmx_servo( ser );
                default:
                        return E( runner_error )
                            << "AUto find for this type of servo not implemented";
        }
}

int main( int argc, char* argv[] )
{

        rclcpp::init( argc, argv );

        auto node = rclcpp::Node::make_shared( "servo_runner" );

        using tmp_type = em::either<
            std::tuple< servo_tester_config, std::optional< serial_connection > >,
            runner_error >;

        get_config< servo_tester_config >( *node, "~" )
            .bind_left( [&]( const servo_tester_config& conf ) -> tmp_type {
                    if ( conf.opt_dev ) {
                            return serial_connection::make( *conf.opt_dev )
                                .convert_left( [&]( serial_connection c ) {
                                        return std::make_tuple( conf, std::make_optional( c ) );
                                } )
                                .construct_right< runner_error >();
                    }
                    return std::make_tuple( conf, std::optional< serial_connection >() );
            } )
            .bind_left(
                em::uncurry | [&]( servo_tester_config                con,
                                   std::optional< serial_connection > opt_ser ) -> tmp_type {
                        if ( opt_ser ) {
                                return find_servo( con.type, *opt_ser )
                                    .convert_left( [&]( servo_id sid ) {
                                            con.opt_id = sid;
                                            return std::make_tuple( con, opt_ser );
                                    } );
                        }
                        con.opt_id = 0;
                        return std::make_tuple( con, opt_ser );
                } )
            .match(
                em::uncurry |
                    [&]( servo_tester_config conf, std::optional< serial_connection > opt_ser ) {
                            servo_runner runner{ conf, opt_ser };

                            if ( conf.opt_source ) {
                                    runner.run_file( *conf.opt_source );
                                    return;
                            }

                            while ( true ) {
                                    runner.tick();
                            }
                    },
                [&]( auto e ) {
                        std::cerr << e;
                } );
}

