// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "schpin_lib/algorithm.h"
#include "schpin_lib/error.h"
#include "schpin_lib/print.h"
#include "schpin_lib/quantity.h"
#include "schpin_lib_exp/algorithm.h"
#include "schpin_lib_exp/bounded.h"
#include "schpin_lib_exp/scaled.h"

#include <filesystem>
#include <magic_enum.hpp>
#include <optional>
#include <rclcpp/rclcpp.hpp>

#pragma once

namespace schpin
{

using config_error = error< struct config_errorTag >;

inline std::ostream& operator<<( std::ostream& os, const config_error& e )
{
        return os << simplified_error{ e };
}

template < typename T >
struct ros_param_serializer
{
        static em::either< T, config_error > load( rclcpp::Node& node, const std::string& prefix )
        {
                auto arg_tuple = T::ros_param_get();

                return std::apply(
                    [&]( auto... item ) {
                            return em::assemble_left_collect_right( item.load( node, prefix )... )
                                .convert_left( [&]( auto value_tpl ) {
                                        return std::apply(
                                            [&]( auto... values ) {  //
                                                    return T{ values... };
                                            },
                                            value_tpl );
                                } )
                                .convert_right( [&]( const auto& errors ) {
                                        return E( config_error ).group_attach( errors );
                                } );
                    },
                    arg_tuple );
        }
};

template < int Len, int Mass, int Time, int Current, int Temp, int Mol, int Li, int Angle, int Byte >
struct ros_param_serializer<
    emlabcpp::physical_quantity< Len, Mass, Time, Current, Temp, Mol, Li, Angle, Byte > >
  : ros_param_serializer< emlabcpp::quantity<
        emlabcpp::physical_quantity< Len, Mass, Time, Current, Temp, Mol, Li, Angle, Byte >,
        float > >
{
};

template < typename T, typename ValueType >
struct ros_param_serializer< em::quantity< T, ValueType > >
{
        static em::either< T, config_error > load( rclcpp::Node& node, const std::string& prefix )
        {
                return ros_param_serializer< ValueType >::load( node, prefix )
                    .template construct_left< T >();
        }
};

template < typename T, T FromVal, T ToVal >
struct ros_param_serializer< bounded< T, FromVal, ToVal > >
{
        static em::either< bounded< T, FromVal, ToVal >, config_error >
        load( rclcpp::Node& node, const std::string& prefix )
        {
                return ros_param_serializer< T >::load( node, prefix )
                    .bind_left(
                        [&]( T val ) -> em::either< bounded< T, FromVal, ToVal >, config_error > {
                                auto opt_val = bounded< T, FromVal, ToVal >::make( val );
                                if ( opt_val ) {
                                        return { *opt_val };
                                }
                                return { E( config_error ) << "Bounded value out of range" };
                        } );
        }
};

template < typename T >
struct ros_param_serializer< std::map< std::string, T > >
{
        static em::either< std::map< std::string, T >, config_error >
        load( rclcpp::Node& node, const std::string& prefix )
        {
                auto params_and_prefixes = node.list_parameters( { prefix }, 1 );

                auto eithers =
                    map_f_to_v( params_and_prefixes.names, [&]( const std::string& name ) {
                            return ros_param_serializer< T >::load( node, prefix + "/" + name )
                                .convert_left( [&]( T item ) {
                                        return std::make_pair( name, item );
                                } );
                    } );

                auto parted = partition_eithers( eithers );
                if ( !parted.right.empty() ) {
                        return {
                            E( config_error ).group_attach( parted.right ) << "Failed to load "
                                                                              "map" };
                }

                std::map< std::string, T > map;
                for ( auto pair : parted.left ) {
                        map.insert( pair );
                }

                return { map };
        }
};

template < typename OutputType, typename SerializationType = OutputType >
struct ros_param_serializer_basic
{
        static em::either< OutputType, config_error >
        load( rclcpp::Node& node, const std::string& prefix )
        {
                if ( !node.has_parameter( prefix ) ) {
                        return { E( config_error ) << "Parameter " << prefix << " is missing" };
                }
                auto param = node.get_parameter( prefix );
                return {
                    static_cast< OutputType >( param.template get_value< SerializationType >() ) };
        }
};

template <>
struct ros_param_serializer< std::string > : ros_param_serializer_basic< std::string >
{
};
template <>
struct ros_param_serializer< double > : ros_param_serializer_basic< double >
{
};
template <>
struct ros_param_serializer< float > : ros_param_serializer_basic< float >
{
};
template <>
struct ros_param_serializer< uint8_t > : ros_param_serializer_basic< uint8_t, int >
{
};
template <>
struct ros_param_serializer< int8_t > : ros_param_serializer_basic< int8_t, int >
{
};
template <>
struct ros_param_serializer< uint16_t > : ros_param_serializer_basic< uint16_t, int >
{
};
template <>
struct ros_param_serializer< int16_t > : ros_param_serializer_basic< int16_t, int >
{
};
template <>
struct ros_param_serializer< uint32_t > : ros_param_serializer_basic< uint32_t, int >
{
};
template <>
struct ros_param_serializer< int32_t > : ros_param_serializer_basic< int32_t, int >
{
};

template < typename Tag, typename T, typename Scale >
struct ros_param_serializer< scaled< Tag, T, Scale > >
{
        static em::either< scaled< Tag, T, Scale >, config_error >
        load( rclcpp::Node& node, const std::string& prefix )
        {
                return ros_param_serializer< float >::load( node, prefix )
                    .convert_left( [&]( float val ) {
                            return scaled< Tag, T, Scale >::from_float( val );
                    } );
        }
};

template <>
struct ros_param_serializer< std::filesystem::path >
{
        static em::either< std::filesystem::path, config_error >
        load( rclcpp::Node& node, const std::string& prefix )
        {
                return ros_param_serializer< std::string >::load( node, prefix )
                    .template construct_left< std::filesystem::path >();
        }
};

template < typename T >
struct ros_param_serializer< std::optional< T > >
{
        static em::either< std::optional< T >, config_error >
        load( rclcpp::Node& node, const std::string& prefix )
        {
                return ros_param_serializer< T >::load( node, prefix )
                    .convert_left( [&]( T val ) {  //
                            return std::make_optional< T >( val );
                    } )
                    .bind_right( [&]( auto ) -> em::either< std::optional< T >, config_error > {
                            return { std::optional< T >{} };
                    } );
        }
};

template < typename T >
struct ros_param_serializer_enum
{
        static em::either< T, config_error > load( rclcpp::Node& node, const std::string& prefix )
        {
                return ros_param_serializer< std::string >::load( node, prefix )
                    .bind_left( [&]( std::string val ) -> em::either< T, config_error > {
                            auto eval = magic_enum::enum_cast< T >( val );
                            if ( !eval.has_value() ) {
                                    return E( config_error )
                                               .group_attach( magic_enum::enum_names< T >() )
                                        << "String " << val
                                        << " is not valid enum value, see list for valid values:";
                            }
                            return eval.value();
                    } );
        }
};

template < typename T >
class ros_param
{
        std::string parameter_name_;
        std::string description_;

public:
        ros_param( std::string parameter_name, std::string description )
          : parameter_name_( std::move( parameter_name ) )
          , description_( std::move( description ) )
        {
        }

        em::either< T, config_error > load( rclcpp::Node& node, const std::string& prefix = "" )
        {
                return ros_param_serializer< T >::load( node, prefix + "/" + parameter_name_ )
                    .convert_right( [&]( config_error e ) -> config_error {
                            return E( config_error ).attach( e )
                                << "Failed to get config parameter: " << term_color::MAGENTA
                                << prefix + parameter_name_ << term_color::RESET
                                << " desc: " << description_;
                    } );
        }
};

template < typename T >
class ros_opt_param
{
        std::string name_;
        std::string description_;
        T           default_;

public:
        ros_opt_param( std::string name, std::string description, T def )
          : name_( std::move( name ) )
          , description_( std::move( description ) )
          , default_( def )
        {
        }

        em::either< T, config_error > load( rclcpp::Node& node, const std::string& prefix = "" )
        {
                return ros_param_serializer< T >::load( node, prefix + "/" + name_ )
                    .bind_right( [&]( const config_error& e ) -> em::either< T, config_error > {  //
                            SCHPIN_INFO_LOG(
                                "config",
                                "Parameter " << name_ << " was not loaded due to: " << e );
                            return default_;
                    } );
        }
};

template < typename Config >
inline em::either< Config, config_error >
get_config( rclcpp::Node& node, const std::string& prefix = "" )
{
        return ros_param_serializer< Config >::load( node, prefix );
}

}  // namespace schpin
