// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "magic_enum.hpp"
#include "schpin_lib/error.h"
#include "schpin_lib/quantity.h"
#include "schpin_lib_exp/bounded.h"
#include "schpin_lib_exp/protocol/message.h"
#include "schpin_lib_exp/protocol/sizeless_message.h"
#include "schpin_lib_exp/scaled.h"

#include <bitset>
#include <emlabcpp/either.h>
#include <emlabcpp/iterators/numeric.h>
#include <variant>

#pragma once

namespace em = emlabcpp;

namespace schpin
{

struct argument_index : em::quantity< argument_index, int16_t >
{
        using em::quantity< argument_index, int16_t >::quantity;

        static std::string get_unit()
        {
                return "";
        }
};

enum protocol_error_enum
{
        INVALID_SIZE_E,
        MISSING_MATCH_E,
        BAD_ARGUMENT_E,
        RUNAWAY_TUPLE_E,
        UNKNOWN_REG_E
};

struct protocol_error_record
{
        protocol_error_enum type;
        std::size_t         byte_index;
};

using protocol_error = error< struct protocol_error_tag >;

inline std::ostream& operator<<( std::ostream& os, const protocol_error_record& rec )
{
        return os << "Error: " << magic_enum::enum_name( rec.type ) << " at: " << rec.byte_index;
}

template < auto ID >
struct cmd_tag
{
        using value_type                  = decltype( ID );
        static constexpr value_type value = ID;
};

template < auto ID >
inline std::ostream& operator<<( std::ostream& os, cmd_tag< ID > )
{
        return os << magic_enum::enum_name( cmd_tag< ID >::value );
}

struct value_tag
{
};

struct error_tag
{
};

enum class protocol_type_id_enum : uint8_t
{
        NONE             = 0,
        UINT8_T          = 1,
        INT8_T           = 2,
        UINT16_T         = 3,
        INT16_T          = 4,
        UINT32_T         = 5,
        INT32_T          = 6,
        UINT64_T         = 7,
        INT64_T          = 8,
        ARRAY            = 9,
        RANGE            = 10,
        FLOAT            = 11,
        MESSAGE          = 12,
        SCALED           = 13,
        BITSET           = 14,
        VARIANT          = 15,
        SIZELESS_MESSAGE = 16
};

namespace protocol
{

        template < typename >
        inline constexpr protocol_type_id_enum type_id_impl = protocol_type_id_enum::NONE;

        template <>
        inline constexpr protocol_type_id_enum type_id_impl< uint8_t > =
            protocol_type_id_enum::UINT8_T;
        template <>
        inline constexpr protocol_type_id_enum type_id_impl< int8_t > =
            protocol_type_id_enum::INT8_T;
        template <>
        inline constexpr protocol_type_id_enum type_id_impl< int16_t > =
            protocol_type_id_enum::INT16_T;
        template <>
        inline constexpr protocol_type_id_enum type_id_impl< uint16_t > =
            protocol_type_id_enum::UINT16_T;
        template <>
        inline constexpr protocol_type_id_enum type_id_impl< int32_t > =
            protocol_type_id_enum::INT32_T;
        template <>
        inline constexpr protocol_type_id_enum type_id_impl< uint32_t > =
            protocol_type_id_enum::UINT32_T;
        template <>
        inline constexpr protocol_type_id_enum type_id_impl< int64_t > =
            protocol_type_id_enum::INT64_T;
        template <>
        inline constexpr protocol_type_id_enum type_id_impl< uint64_t > =
            protocol_type_id_enum::UINT64_T;
        template <>
        inline constexpr protocol_type_id_enum type_id_impl< float > = protocol_type_id_enum::FLOAT;
        template < typename MsgCounterType, std::size_t N, typename Owner >
        inline constexpr protocol_type_id_enum
            type_id_impl< prot_message< MsgCounterType, N, Owner > > =
                protocol_type_id_enum::MESSAGE;
        template < typename T, T Min, T Max >
        inline constexpr protocol_type_id_enum type_id_impl< bounded< T, Min, Max > > =
            protocol_type_id_enum::RANGE;
        template < typename T, std::size_t N >
        inline constexpr protocol_type_id_enum type_id_impl< std::array< T, N > > =
            protocol_type_id_enum::ARRAY;
        template < typename Tag, typename T, typename Scale >
        inline constexpr protocol_type_id_enum type_id_impl< scaled< Tag, T, Scale > > =
            protocol_type_id_enum::SCALED;
        template < std::size_t N >
        inline constexpr protocol_type_id_enum type_id_impl< std::bitset< N > > =
            protocol_type_id_enum::BITSET;
        template < typename... Ts >
        inline constexpr protocol_type_id_enum type_id_impl< std::variant< Ts... > > =
            protocol_type_id_enum::VARIANT;
        template < std::size_t N >
        inline constexpr protocol_type_id_enum type_id_impl< prot_sizeless_message< N > > =
            protocol_type_id_enum::SIZELESS_MESSAGE;

}  // namespace protocol

template < typename T >
constexpr uint8_t protocol_type_id_v = static_cast< uint8_t >( protocol::type_id_impl< T > );

}  // namespace schpin
