// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include <schpin_lib_exp/protocol.h>

#pragma once

namespace schpin
{

template < auto KeyV, typename T >
struct prot_conf_parameter_s
{
        T& param;

        static constexpr auto key = KeyV;
        using cmd                 = typename command< KeyV >::template with_args< T >;
        using key_type            = decltype( KeyV );

        explicit prot_conf_parameter_s( T& para )
          : param( para )
        {
        }
};

template < auto Key, typename T >
prot_conf_parameter_s< Key, T > prot_conf_parameter( T* val )
{
        return prot_conf_parameter_s< Key, T >{ *val };
}

enum prot_config_mc_enum : uint8_t
{
        PROT_CONFIG_SET,
        PROT_CONFIG_GET,
        PROT_CONFIG_ENUMERATE
};
enum prot_config_cm_enum : uint8_t
{
        PROT_CONFIG_RESPONSE,
        PROT_CONFIG_ERROR,
        PROT_CONFIG_ENUMERATE_FINISHED
};

enum prot_config_error_enum
{
        PROT_CONFIG_KEY_MISSING,
        PROT_CONFIG_VALUE_MISMATCH,
        PROT_CONFIG_MISSING_LOAD_KEY
};

template < typename Config >
struct prot_config_types
{
        template < typename... >
        struct extractor;

        template < typename... Items >
        struct extractor< std::tuple< Items... > >
        {
                using parameters_group   = command_group<>::with_commands< typename Items::cmd... >;
                using parameters_message = typename parameters_group::message_t;

                static_assert( em::are_same_v< typename Items::key_type... > );

                using key_type =
                    typename std::tuple_element_t< 0, std::tuple< Items... > >::key_type;
        };

        using bind_tpl = decltype( std::declval< Config >().bind_prot_config() );

        static_assert( em::is_std_tuple_v< bind_tpl > );

        using parameters_group   = typename extractor< bind_tpl >::parameters_group;
        using parameters_message = typename parameters_group::message_t;
        using parameters_variant = typename parameters_group::command_variant;
        using key_type           = typename extractor< bind_tpl >::key_type;

        using master_config_group = command_group<>::with_commands<
            command< PROT_CONFIG_SET >::with_promoted_args< parameters_message >,
            command< PROT_CONFIG_GET >::with_args< key_type >,
            command< PROT_CONFIG_ENUMERATE > >;
        using master_config_message         = typename master_config_group::message_t;
        using master_config_command_variant = typename master_config_group::command_variant;

        using config_master_group = command_group<>::with_commands<
            command< PROT_CONFIG_RESPONSE >::with_promoted_args< parameters_message >,
            command< PROT_CONFIG_ERROR >::with_args< prot_config_error_enum >,
            command< PROT_CONFIG_ENUMERATE_FINISHED > >;
        using config_master_message         = typename config_master_group::message_t;
        using config_master_command_variant = typename config_master_group::command_variant;
};

template < typename Config >
class prot_config_loader
{
        enum state_enum
        {
                NOTHING,
                ENUMERATING,
                LOADED
        };

        Config     config_;
        state_enum state_ = NOTHING;

        using types = prot_config_types< Config >;

      public:
        using key_type           = typename types::key_type;
        using parameters_group   = typename types::parameters_group;
        using parameters_message = typename types::parameters_message;
        using parameters_variant = typename types::parameters_variant;

        using master_config_group           = typename types::master_config_group;
        using master_config_message         = typename master_config_group::message_t;
        using master_config_command_variant = typename master_config_group::command_variant;

        using config_master_group           = typename types::config_master_group;
        using config_master_message         = typename config_master_group::message_t;
        using config_master_command_variant = typename config_master_group::command_variant;

        prot_config_loader() = default;
        explicit prot_config_loader( Config c )
          : config_( c )
        {
        }

        Config& operator*()
        {
                return config_;
        }

        const Config& operator*() const
        {
                return config_;
        }

        Config* operator->()
        {
                return &config_;
        }
        const Config* operator->() const
        {
                return &config_;
        }

        bool has_loaded()
        {
                return state_ != NOTHING;
        }

        template < typename Visitor >
        void load_config( Visitor&& vis )
        {
                state_ = ENUMERATING;
                send_msg< PROT_CONFIG_ENUMERATE >( vis );
        }

        template < typename Visitor >
        void on_servo_config_message( config_master_command_variant var, Visitor&& v )
        {
                var << tuple_match(
                    [&]( cmd_tag< PROT_CONFIG_RESPONSE >, parameters_variant pvar ) {
                            if ( state_ != ENUMERATING ) {
                                    return;
                            }
                            pvar << tuple_match( [&]( auto cmd_tag, auto ) {
                                    send_set_value< decltype( cmd_tag )::value >( v );
                            } );
                    },
                    [&]( cmd_tag< PROT_CONFIG_ERROR >, prot_config_error_enum ) {
                            // TODO: proper error handling
                    },
                    [&]( cmd_tag< PROT_CONFIG_ENUMERATE_FINISHED > ) {
                            state_ = LOADED;
                    } );
        }

      private:
        template < prot_config_mc_enum K, typename Visitor, typename... Args >
        void send_msg( Visitor&& v, Args... args )
        {
                v( master_config_group::template create< K >( args... ) );
        }

        template < key_type K, typename Visitor >
        void send_set_value( Visitor&& v )
        {
                em::for_each( config_.bind_prot_config(), [&]( auto param_bind ) {
                        if constexpr ( decltype( param_bind )::key == K ) {
                                send_msg< PROT_CONFIG_SET >(
                                    v, parameters_group::template create< K >( param_bind.param ) );
                        }
                } );
        }
};

// TODO: split into a class for master->config and config->master separately
template < typename Config >
class prot_config
{
        Config config_;

        using types = prot_config_types< Config >;

      public:
        using key_type           = typename types::key_type;
        using parameters_group   = typename types::parameters_group;
        using parameters_message = typename types::parameters_message;
        using parameters_variant = typename types::parameters_variant;

        using master_config_group           = typename types::master_config_group;
        using master_config_message         = typename master_config_group::message_t;
        using master_config_command_variant = typename master_config_group::command_variant;

        using config_master_group           = typename types::config_master_group;
        using config_master_message         = typename config_master_group::message_t;
        using config_master_command_variant = typename config_master_group::command_variant;

        const Config* operator->() const
        {
                return &config_;
        }

        template < typename Visitor >
        void on_master_config_msg( master_config_command_variant var, Visitor&& v )
        {
                var << tuple_match(
                    [&]( cmd_tag< PROT_CONFIG_SET >, parameters_variant var ) {  //
                            set_param( var, v );
                    },
                    [&]( cmd_tag< PROT_CONFIG_GET >, key_type k ) {  //
                            reply_param( k, v );
                    },
                    [&]( cmd_tag< PROT_CONFIG_ENUMERATE > ) {  //
                            enumerate_params( v );
                    } );
        }

      private:
        template < typename Visitor >
        void set_param( parameters_variant var, Visitor&& )
        {
                var << tuple_match( [&]( auto param_tag, auto value ) {
                        em::for_each( config_.bind_prot_config(), [&]( auto param_bind ) {
                                if constexpr ( decltype( param_bind )::key ==
                                               std::decay_t< decltype( param_tag ) >::value ) {
                                        param_bind.param = value;
                                }
                        } );
                } );
        }

        template < typename Visitor >
        void reply_param( key_type k, Visitor&& v )
        {
                em::for_each( config_.bind_prot_config(), [&]( auto param_bind ) {
                        if ( decltype( param_bind )::key == k ) {
                                send_reply< decltype( param_bind )::key >( param_bind.param, v );
                        }
                } );
        }

        template < typename Visitor >
        void enumerate_params( Visitor&& v )
        {
                em::for_each( config_.bind_prot_config(), [&]( auto param_bind ) {
                        send_reply< decltype( param_bind )::key >( param_bind.param, v );
                } );
                v( config_master_group::template create< PROT_CONFIG_ENUMERATE_FINISHED >() );
        }

        template < key_type K, typename T, typename Visitor >
        void send_reply( T val, Visitor&& v )
        {
                v( config_master_group::template create< PROT_CONFIG_RESPONSE >(
                    parameters_group::template create< K >( val ) ) );
        }
};

}  // namespace schpin
