// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include <schpin_lib_exp/protocol/base.h>

#pragma once

namespace schpin
{

template < auto ID, typename T >
struct protocol_reg
{
        static constexpr auto id = ID;
        using value_type         = T;

        using prot_item                   = protocol_item< value_type >;
        static constexpr std::size_t size = prot_item::size;

        T value;
};

template < typename... >
struct protocol_reg_storage;

template < auto ID >
struct protocol_reg_storage_tag
{
};

template <>
struct protocol_reg_storage<>
{

        template < auto ID >
        static constexpr bool contains_id = false;

        template < auto ID >
        using value_t = void;

        static constexpr std::size_t get_size( std::size_t )
        {
                // maybe assert that this does not happen?
                return 0;
        }

        void get_val() const
        {
        }

        void set_val() const
        {
        }

        template < typename T, typename Buffer >
        opt_error< protocol_error_record > set_raw_val( T, const Buffer& )
        {
                return { protocol_error_record{ UNKNOWN_REG_E, 0 } };
        }

        template < typename Buffer, typename Enum >
        Buffer get_buffered_val( Enum ) const
        {
                // todo: this should be an error :/
                return Buffer{};
        }
};

template < typename Reg, typename... Regs >
struct protocol_reg_storage< Reg, Regs... > : public protocol_reg_storage< Regs... >
{
        using sub       = protocol_reg_storage< Regs... >;
        using enum_type = decltype( Reg::id );

        Reg reg_;

        template < auto ID >
        static constexpr bool contains_id = ID == Reg::id ? true : sub::template contains_id< ID >;

        template < auto ID >
        using value_t = std::conditional_t<
            ID == Reg::id,
            typename Reg::value_type,
            typename sub::template value_t< ID > >;

        protocol_reg_storage() = default;
        protocol_reg_storage( typename Reg::value_type val, typename Regs::value_type... subvals )
          : sub( subvals... )
          , reg_{ val }
        {
        }

        static constexpr std::size_t get_size( std::size_t i )
        {
                if ( i == 0 ) {
                        return Reg::size;
                }
                return sub::get_size( i - 1 );
        }

        static constexpr std::size_t get_size( protocol_reg_storage_tag< Reg::id > )
        {
                return Reg::size;
        }
        using sub::get_size;

        typename Reg::value_type get_val( protocol_reg_storage_tag< Reg::id > ) const
        {
                return reg_.value;
        }
        using sub::get_val;

        template < typename Buffer >
        Buffer get_buffered_val( enum_type reg_val ) const
        {
                if ( reg_val != Reg::id ) {
                        return sub::template get_buffered_val< Buffer >( reg_val );
                }
                // TODO this is bad idea, the view may not be big enough :/
                std::array< uint8_t, Reg::size > buff;
                std::size_t used = protocol_item< typename Reg::value_type >::serialize_at(
                    make_static_view( buff ), reg_.value );
                return Buffer{ buff, used };
        }

        template < typename Buffer >
        opt_error< protocol_error_record > set_raw_val( enum_type reg, const Buffer& buff )
        {
                opt_error< protocol_error_record > res;
                if ( reg != Reg::id ) {
                        return sub::set_raw_val( reg, buff );
                }
                protocol_item< typename Reg::value_type >::deserialize(
                    buff,
                    match(
                        [&]( value_tag, auto val ) {
                                reg_.value = val;
                        },
                        [&]( error_tag, protocol_error_record rec ) {
                                res = rec;
                        } ) );

                return res;
        }

        void set_val( protocol_reg_storage_tag< Reg::id >, typename Reg::value_type val )
        {
                reg_.value = val;
        }
        using sub::set_val;
};

template < typename Enum, typename... Regs >
class protocol_map
{
        using storage = protocol_reg_storage< Regs... >;

        storage stor_;

public:
        static constexpr std::size_t                    item_count = sizeof...( Regs );
        static constexpr std::size_t                    max_size   = em::max( { Regs::size... } );
        static constexpr std::array< Enum, item_count > ids        = { Regs::id... };

        template < auto ID >
        using reg_value_t = typename storage::template value_t< ID >;

        template < auto ID >
        static constexpr bool contains_id = storage::template contains_id< ID >;

        template < auto ID >
        struct tag
        {
                static constexpr auto reg = ID;
                using value_type          = reg_value_t< ID >;
        };

        protocol_map() = default;
        protocol_map( typename Regs::value_type... def_vals )
          : stor_( def_vals... )
        {
        }

        template < auto ID >
        static constexpr std::size_t get_size()
        {
                static_assert( contains_id< ID > );
                return storage::get_size( protocol_reg_storage_tag< ID >{} );
        }

        static constexpr std::size_t get_size( std::size_t i )
        {
                return storage::get_size( i );
        }

        template < auto ID >
        reg_value_t< ID > get_val() const
        {
                static_assert( contains_id< ID > );
                return stor_.get_val( protocol_reg_storage_tag< ID >{} );
        }

        template < typename Buffer >
        Buffer get_buffered_val( Enum reg_val ) const
        {
                return stor_.template get_buffered_val< Buffer >( reg_val );
        }

        template < typename Buffer >
        opt_error< protocol_error_record > set_val( Enum reg, const Buffer& buff )
        {
                // TODO: proper error handling
                return stor_.set_raw_val( reg, buff );
        }

        template < auto ID >
        void set_val( reg_value_t< ID > val )
        {
                static_assert( contains_id< ID > );
                stor_.set_val( protocol_reg_storage_tag< ID >{}, val );
        }
};

template < typename T, typename UnaryFunction >
inline void for_each_reg( T&&, UnaryFunction&& f )
{
        using MapT = std::decay_t< T >;
        em::for_each( em::make_sequence_tuple_t< MapT::item_count >{}, [&]( auto ci ) {
                static constexpr std::size_t i   = decltype( ci )::value;
                static constexpr auto        reg = MapT::ids[i];
                f( typename MapT::tag< reg >{} );
        } );
}

template < typename Enum, typename... Regs >
inline std::ostream& operator<<( std::ostream& os, const protocol_map< Enum, Regs... >& mp )
{
        for_each_reg( mp, [&]( auto tok ) {
                static constexpr Enum reg = decltype( tok )::reg;

                os << magic_enum::enum_name( reg ) << ": " << mp.template get_val< reg >() << "\n";
        } );

        return os;
}

template < typename Enum, typename... Regs >
inline void from_json( const nlohmann::json& j, protocol_map< Enum, Regs... >& mp )
{
        for_each_reg( mp, [&]( auto tok ) {
                static constexpr Enum reg = decltype( tok )::reg;
                using T                   = typename decltype( tok )::value_type;

                std::string name{ magic_enum::enum_name( reg ) };
                mp.template set_val< reg >( j[name].template get< T >() );
        } );
}

template < typename Enum, typename... Regs >
inline void to_json( nlohmann::json& j, const protocol_map< Enum, Regs... >& mp )
{
        for_each_reg( mp, [&]( auto tok ) {
                static constexpr Enum reg = decltype( tok )::reg;

                std::string name{ magic_enum::enum_name( reg ) };
                j[name] = mp.template get_val< reg >();
        } );
}

}  // namespace schpin
