// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "environment.h"

#include "base/triangle.h"
#include "octree/insert.h"

namespace schpin
{

class environment_surface_visitor
  : public i_oc_simple_visitor< environment_oc_tree_content, world_frame >
{
        const env_oc_tree& tree_;

public:
        environment_surface_visitor( env_oc_tree& tree )
          : tree_( tree )
        {
        }

        oc_node_control access_node( const env_proxy_node& /*unused*/ ) const final
        {
                return oc_node_control::FOLLOW_CHILDREN;
        }

        oc_locked_leaf_control access_locked_leaf( const env_proxy_node& /*unused*/ ) const final
        {
                return oc_locked_leaf_control::EXPAND_CHILDREN;
        }

        void access_bottom_leaf( const env_proxy_node& node, environment_oc_tree_content& content )
            const final
        {

                // verify that actual node holds 'air'
                if ( ( *node )->get_occupancy() >= ( env_min_occupancy + env_prunning_diff ) ) {
                        return;
                }

                point< 3, world_frame > desired_pos =
                    node.get_position() + vec< 3 >{ 0, 0, -node.get_edge_length() };

                auto opt_below_node = find_leaf_node( tree_, desired_pos );

                if ( !opt_below_node ) {
                        return;
                }

                const env_proxy_node& below_node = *opt_below_node;

                // verify that node below holds ground
                if ( ( *below_node )->get_occupancy() < env_max_occupancy - env_prunning_diff ) {
                        return;
                }

                content.set_surface( true );
        }
};

em::either< environment, env_error > build_environment( const env_config& config )
{
        SCHPIN_INFO_LOG( "env", "Building environment" );
        SCHPIN_INFO_LOG( "env", "Default occ " << config.default_occ );
        SCHPIN_INFO_LOG( "env", "Min edge length " << config.min_edge_length );
        for ( const std::filesystem::path& p : config.paths ) {
                SCHPIN_INFO_LOG( "env", "Path " << p );
        }

        using visitor = single_collision_visitor< fast_collision_checker<
            triangle_collision_query< 3, world_frame >,
            triangle_fast_collision_query< world_frame >,
            environment_oc_tree_content > >;
        /*
using Visitor = single_collision_visitor<
    FullCollisionChecker< triangle_collision_query< 3, world_frame >, environment_oc_tree_content >
>;*/

        std::vector< parse_error > errors;
        env_oc_tree                tree{
            environment_oc_tree_content( config.default_occ, false ), config.min_edge_length };

        std::size_t c = 0;

        auto handle_triangles_f = [&](
                                      const std::vector< triangle< 3, world_frame > >& triangles ) {
                auto bound_cube = dimensional_min_max_elem< point< 3, world_frame > >(
                    em::range( triangles.size() * 3 ),
                    [&]( std::size_t i ) {  //
                            return triangles[i / 3][i % 3];
                    } );
                SCHPIN_INFO_LOG( "env", "Triangle area: " << bound_cube );
                for ( const triangle< 3, world_frame >& t : triangles ) {
                        ++c;
                        SCHPIN_INFO_LOG(
                            "env", "Triangle processed count: " << c << "/" << triangles.size() );
                        visitor v{ t, occupancy_token::FULL };
                        tree.expand( t );
                        tree.modify( v );
                }
        };

        for ( const std::filesystem::path& p : config.paths ) {
                parse_triangles< world_frame >( p ).match(
                    handle_triangles_f,
                    [&]( const parse_error& error ) {  //
                            errors.push_back( error );
                    } );
        }

        if ( !errors.empty() ) {
                return {
                    E( env_error ).group_attach( errors )
                    << "Failed to build environment due to errors in parsing" };
        }

        environment_surface_visitor surface_visitor{ tree };
        tree.modify( surface_visitor );
        SCHPIN_INFO_LOG( "env", "Environment size: " << get_node_count( tree.get_proxy_root() ) );
        auto opt_space = get_used_space( tree.get_proxy_root(), [&]( const auto& pnode ) {
                return ( *pnode )->get_occupancy() > env_max_occupancy / 2;
        } );
        if ( opt_space ) {
                SCHPIN_INFO_LOG( "env", "Environment used space: " << *opt_space );
        }

        return { environment( std::move( tree ) ) };
}

}  // namespace schpin
