// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "collission.h"
#include "octree/interfaces.h"
#include "octree/occupancy.h"
#include "octree/proxy_node.h"
#include "octree/tree.h"

#pragma once

namespace schpin
{

template < typename T, typename Content >
class full_collision_checker : public i_oc_collision_checker< Content, typename T::tag >
{
      public:
        using tag = typename T::tag;

      private:
        const T item_query_;

      public:
        template < typename ItemType >
        explicit full_collision_checker( const ItemType& item )
          : item_query_( item )
        {
        }
        constexpr bool inner_collission(
            const oc_node_full_collision_query< Content, tag >& node_query ) const final
        {
                return collides< 3 >( item_query_, node_query );
        }
        constexpr bool bottom_collission(
            const oc_node_full_collision_query< Content, tag >& node_query ) const final
        {
                return collides< 3 >( item_query_, node_query );
        }
};

template < typename T, typename TFast, typename Content >
class fast_collision_checker : public i_oc_collision_checker< Content, typename T::tag >
{
      public:
        using tag = typename T::tag;

      private:
        const T     item_query_;
        const TFast item_fast_query_;

      public:
        template < typename ItemType >
        explicit fast_collision_checker( const ItemType& item )
          : item_query_( item )
          , item_fast_query_( item )
        {
        }

        constexpr bool inner_collission(
            const oc_node_full_collision_query< Content, tag >& node_query ) const final
        {
                oc_node_fast_collision_query fast_q{ *node_query };
                return fast_collides< 3 >( item_fast_query_, fast_q );
        }

        constexpr bool bottom_collission(
            const oc_node_full_collision_query< Content, tag >& node_query ) const final
        {
                return collides< 3 >( item_query_, node_query );
        }
};

template < typename Checker >
class single_collision_visitor
  : public i_oc_simple_visitor< typename Checker::node_content, typename Checker::tag >
{
      public:
        using node_content = typename Checker::node_content;
        using token        = typename node_content::token;
        using tag          = typename Checker::tag;

      private:
        static_assert(
            std::is_base_of< i_oc_collision_checker< node_content, tag >, Checker >::value );
        Checker checker_;
        token   token_;

      public:
        template < typename Item >
        single_collision_visitor( const Item& item, token token )
          : checker_( item )
          , token_( token )
        {
        }
        constexpr oc_node_control
        access_node( const oc_proxy_node< node_content, tag >& node ) const final
        {
                oc_node_full_collision_query node_query( node );
                if ( checker_.inner_collission( node_query ) ) {
                        return oc_node_control::FOLLOW_CHILDREN;
                }
                return oc_node_control::END;
        }
        constexpr oc_locked_leaf_control
        access_locked_leaf( const oc_proxy_node< node_content, tag >& node ) const final
        {
                oc_node_full_collision_query node_query( node );
                if ( checker_.inner_collission( node_query ) ) {
                        return oc_locked_leaf_control::EXPAND_CHILDREN;
                }
                return oc_locked_leaf_control::END;
        }

        void access_bottom_leaf( const oc_proxy_node< node_content, tag >& node,
                                 node_content&                             content ) const final
        {
                oc_node_full_collision_query node_query( node );
                if ( checker_.bottom_collission( node_query ) ) {
                        content.update( token_ );
                }
        }
};

template < typename Checker, unsigned BatchSize >
class batch_collision_visitor
  : i_oc_visitor< std::bitset< BatchSize >, typename Checker::node_content, typename Checker::tag >
{
      public:
        using node_content = typename Checker::node_content;
        using tag          = typename Checker::tag;
        using info_type    = std::bitset< BatchSize >;
        using token        = typename node_content::token;

      private:
        static_assert(
            std::is_base_of< i_oc_collision_checker< node_content, tag >, Checker >::value );

        std::vector< Checker > checkers_{};
        info_type              basic_selection_{ 0 };
        token                  token_;

      public:
        template < typename Iterator >
        batch_collision_visitor( batch< Iterator > items, token token )
          : token_( token )
        {
                std::size_t i = 0;
                for ( const auto& item : items ) {
                        checkers_.emplace_back( item );
                        basic_selection_[i] = 1;
                        ++i;
                }
        }

        constexpr std::bitset< BatchSize > get_basic_selection() const
        {
                return basic_selection_;
        }

        constexpr std::pair< oc_node_control, info_type >
        access_node( const oc_proxy_node< node_content, tag >& node,
                     const info_type&                          parent_info ) const final
        {
                info_type                    info = parent_info;
                oc_node_full_collision_query node_query( node );
                for ( std::size_t i : em::range( BatchSize ) ) {
                        if ( !info[i] ) {
                                continue;
                        }
                        if ( !checkers_[i].inner_collission( node_query ) ) {
                                info[i] = false;
                        }
                }
                if ( info.any() ) {
                        return { oc_node_control::FOLLOW_CHILDREN, info };
                }

                return { oc_node_control::END, info };
        }

        constexpr std::pair< oc_locked_leaf_control, info_type >
        access_locked_leaf( const oc_proxy_node< node_content, tag >& node,
                            const info_type&                          parent_info ) const final
        {
                std::pair< oc_node_control, info_type > res = access_node( node, parent_info );
                switch ( res.first ) {
                        case oc_node_control::FOLLOW_CHILDREN:
                                return { oc_locked_leaf_control::EXPAND_CHILDREN, res.second };
                        case oc_node_control::END:
                                return { oc_locked_leaf_control::END, res.second };
                }
                return { oc_locked_leaf_control::END, res.second };
        }

        void access_bottom_leaf( const oc_proxy_node< node_content, tag >& node,
                                 const info_type&                          info,
                                 node_content&                             content ) const final
        {
                oc_node_full_collision_query node_query( node );

                for ( std::size_t i : em::range( BatchSize ) ) {
                        if ( info[i] && checkers_[i].bottom_collission( node_query ) ) {
                                content.update( token_ );
                        }
                }
        }
};
}  // namespace schpin
