// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "octree/tree.h"
#include "schpin_lib/geom/point/serialization.h"
#include "schpin_lib/json.h"
#include "schpin_lib/parser.h"
#include "schpin_lib/quantity.h"

#include <nlohmann/json.hpp>

#pragma once

namespace schpin
{

template < typename InsertionVisitor, typename Tag >
oc_tree< typename InsertionVisitor::node_content, Tag >
build_oc_tree_iterative( const std::vector< triangle< 3, Tag > >&      triangles,
                         const typename InsertionVisitor::node_content content,
                         const em::length                              min_resolution,
                         const typename InsertionVisitor::Token        token )
{
        using node_content = typename InsertionVisitor::node_content;
        oc_tree< node_content, Tag > res{ content, min_resolution };

        for ( const triangle< 3, Tag >& tri : triangles ) {
                res.expand( tri );

                InsertionVisitor visitor( tri, token );

                res.modify( visitor );
        }

        return res;
}

}  // namespace schpin

template < typename Content >
struct nlohmann::adl_serializer< schpin::oc_node< Content > >
{
        static schpin::oc_node< Content > from_json( const nlohmann::json& j )
        {

                using array = std::array< schpin::oc_node< Content >, 8 >;

                nlohmann::json j_children = j.at( "children" );

                std::unique_ptr< array > ptr;

                if ( !j_children.empty() ) {
                        auto vec = j_children.get< std::vector< schpin::oc_node< Content > > >();
                        // TODO(squirrel): exception should be raised in case vec.size() != 8
                        // TODO: this is shitty as hell
                        ptr = std::make_unique< array >( array{ std::move( vec.at( 0 ) ),
                                                                std::move( vec.at( 1 ) ),
                                                                std::move( vec.at( 2 ) ),
                                                                std::move( vec.at( 3 ) ),
                                                                std::move( vec.at( 4 ) ),
                                                                std::move( vec.at( 5 ) ),
                                                                std::move( vec.at( 6 ) ),
                                                                std::move( vec.at( 7 ) ) } );
                }

                return schpin::oc_node< Content >( std::move( ptr ),
                                                   j.at( "content" ).get< Content >() );
        }

        static void to_json( nlohmann::json& j, const schpin::oc_node< Content >& node )
        {
                j["content"]  = *node;
                j["children"] = node.get_children();
        }
};

template < typename Content, typename Tag >
struct nlohmann::adl_serializer< schpin::oc_tree< Content, Tag > >
{
        static schpin::oc_tree< Content, Tag > from_json( const nlohmann::json& j )
        {
                auto root = j.at( "root" ).get< schpin::oc_node< Content > >();
                schpin::oc_node_context< Tag > context{
                    j.at( "depth" ).get< unsigned >(),
                    j.at( "max_depth" ).get< unsigned >(),
                    j.at( "length" ).get< emlabcpp::length >(),
                    j.at( "position" ).get< schpin::point< 3, Tag > >() };

                schpin::oc_tree< Content, Tag > tree{
                    std::move( root ),
                    std::move( context ),
                    j.at( "default_content" ).get< Content >(),
                };

                SCHPIN_ASSERT( validate_octree( tree ) );

                return tree;
        }

        static void to_json( nlohmann::json& j, const schpin::oc_tree< Content, Tag >& tree )
        {
                const schpin::oc_node_context< Tag >& root_context = tree.get_root_context();

                j["root"]            = tree.get_bare_root();
                j["default_content"] = tree.get_default_content();
                j["depth"]           = root_context.depth;
                j["max_depth"]       = root_context.max_depth;
                j["length"]          = root_context.edge_length;
                j["position"]        = root_context.position;
        }
};
