// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "base.h"
#include "base/physics.h"
#include "schpin_lib/geom/point.h"

#include <utility>

#pragma once

namespace schpin
{

template < typename Tag >
struct oc_node_context
{
        unsigned        depth;
        unsigned        max_depth;
        em::length      edge_length;
        point< 3, Tag > position;
};

template < typename Tag >
inline std::ostream& operator<<( std::ostream& os, const oc_node_context< Tag >& cont )
{
        os << "Depth: " << cont.depth << " MaxDepth: " << cont.max_depth
           << " Edgelength: " << cont.edge_length << " Position: " << cont.position;
        return os;
}

template < typename Tag >
constexpr bool operator==( const oc_node_context< Tag >& lh, const oc_node_context< Tag >& rh )
{
        return lh.depth == rh.depth && lh.max_depth == rh.max_depth &&
               lh.edge_length == rh.edge_length && lh.position == rh.position;
}

template < typename Tag >
constexpr bool operator<( const oc_node_context< Tag >& lh, const oc_node_context< Tag >& rh )
{
        return std::tie( lh.depth, lh.max_depth, lh.edge_length, lh.position, lh.edge_length ) <
               std::tie( rh.depth, rh.max_depth, rh.edge_length, rh.position, rh.edge_length );
}

template < typename Tag >
inline oc_node_context< Tag > upscale_context( const oc_node_context< Tag >& context,
                                               oc_child_index                actual_index )
{
        unsigned new_depth     = context.depth == 0 ? 0 : context.depth - 1;
        unsigned new_max_depth = context.max_depth + ( context.depth == 0 ? 1 : 0 );
        return oc_node_context< Tag >{
            new_depth,
            new_max_depth,
            context.edge_length * 2.f,
            cast_to_point< Tag >(
                context.position -
                get_oc_child_offset< Tag >( actual_index, context.edge_length * 2.f ) )};
}

template < typename Tag >
inline oc_node_context< Tag > downscale_context( const oc_node_context< Tag >& context,
                                                 oc_child_index                actual_index )
{
        return oc_node_context< Tag >{context.depth + 1,
                                      context.max_depth,
                                      context.edge_length / 2.f,
                                      context.position + cast_to_vec( get_oc_child_offset< Tag >(
                                                             actual_index, context.edge_length ) )};
}
}  // namespace schpin

template < typename Tag >
struct std::hash< schpin::oc_node_context< Tag > >
{
        std::size_t operator()( const schpin::oc_node_context< Tag >& node )
        {
                std::stringstream ss;
                ss << node;
                return std::hash< std::string >()( ss.str() );
        }
};
