// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "collission_interfaces.h"
#include "octree/node.h"
#include "octree/node_context.h"
#include "schpin_lib/geom/pose.h"
#include "util.h"

#include <utility>

#pragma once

namespace schpin
{

template < typename NodeContent, typename Tag >
class oc_tree;

template < typename Content, typename Tag >
class oc_proxy_node
{

      public:
        const oc_node< Content >& operator*() const
        {
                return bare_node_;
        }
        const oc_node< Content >* operator->() const
        {
                return &bare_node_;
        }

        template < typename UnaryFunction >
        void modify_content( UnaryFunction f )
        {
                f( *bare_node_ );
        }

        constexpr std::size_t size() const
        {
                return bare_node_.size();
        }

        constexpr bool has_children() const
        {
                return bare_node_.has_children();
        }

        constexpr unsigned get_max_depth() const
        {
                return context_.max_depth;
        }

        constexpr unsigned get_depth() const
        {
                return context_.depth;
        }

        constexpr em::length get_edge_length() const
        {
                return context_.edge_length;
        }

        constexpr const point< 3, Tag >& get_position() const
        {
                return context_.position;
        }

        constexpr bool is_leaf() const
        {
                return bare_node_.is_leaf();
        }

        // method possible to execute only some times? sucks
        constexpr void init_children()
        {
                bare_node_.init_children();
        }

        void delete_children()
        {
                bare_node_.delete_children();
        }

        oc_proxy_node get_child( uint8_t index )
        {
                SCHPIN_ASSERT( index < bare_node_.size() );
                return bare_node_.template overlay_children< oc_proxy_node >(
                    index, downscale_context( context_, index ) );
        }

        // TODO: this is definetly not correct
        oc_proxy_node get_child( uint8_t index ) const
        {
                SCHPIN_ASSERT( index < bare_node_.size() );

                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-const-cast)
                return const_cast< oc_proxy_node* >( this )->get_child( index );
        }

        std::array< oc_proxy_node, 8 > generate_children()
        {
                SCHPIN_ASSERT( bare_node_.has_children() );
                return { get_child( 0 ),
                         get_child( 1 ),
                         get_child( 2 ),
                         get_child( 3 ),
                         get_child( 4 ),
                         get_child( 5 ),
                         get_child( 6 ),
                         get_child( 7 ) };
        }
        std::array< const oc_proxy_node, 8 > generate_children() const
        {
                SCHPIN_ASSERT( bare_node_.has_children() );
                return { get_child( 0 ),
                         get_child( 1 ),
                         get_child( 2 ),
                         get_child( 3 ),
                         get_child( 4 ),
                         get_child( 5 ),
                         get_child( 6 ),
                         get_child( 7 ) };
        }

        void try_prunning()
        {
                if ( !has_children() ) {
                        return;
                }
                if ( em::any_of( bare_node_.get_children(),
                                 [&]( const oc_node< Content >& node ) {  //
                                         return node.has_children();
                                 } ) ) {
                        return;
                }
                const oc_node< Content >& cref = bare_node_;
                if ( bare_node_->prunnable( cref ) ) {
                        bare_node_->set_prunned( cref );
                        bare_node_.delete_children();
                }
        }

      private:
        friend oc_tree< Content, Tag >;
        friend oc_node< Content >;

        oc_proxy_node( oc_node< Content >& node, oc_node_context< Tag > context )
          : bare_node_( node )
          , context_( std::move( context ) )
        {
        }

        oc_node< Content >&          bare_node_;
        const oc_node_context< Tag > context_;
};

template < typename NodeContent, typename Tag >
std::ostream& operator<<( std::ostream& os, const oc_proxy_node< NodeContent, Tag >& node )
{
        os << " oc_proxy_node pos: " << node.get_position() << " size: " << node.size()
           << " depth: " << node.get_depth() << "/" << node.get_max_depth()
           << " edge: " << node.get_edge_length();
        return os;
}

template < typename NodeContent, typename Tag >
constexpr bool contains( const oc_proxy_node< NodeContent, Tag >& node,
                         const point< 3, Tag >&                   target )
{
        const point< 3, Tag >& pos = node.get_position();
        em::length             l   = node.get_edge_length() / 2.f;

        return em::all_of( em::range( 3u ), [&]( std::size_t i ) {  //
                return pos[i] - *l <= target[i] && target[i] <= pos[i] + *l;
        } );
}

template < typename NodeContent, typename Tag >
constexpr uint64_t get_node_count( const oc_proxy_node< NodeContent, Tag >& node )
{
        return get_node_count( *node );
}

template < typename NodeContent, typename Tag, typename Predicate >
constexpr uint64_t get_node_count( const oc_proxy_node< NodeContent, Tag >& node, Predicate&& pred )
{
        return get_node_count( *node, std::forward< Predicate >( pred ) );
}

template < typename NodeContent, typename Tag >
constexpr uint64_t get_tree_height( const oc_proxy_node< NodeContent, Tag >& node )
{
        return get_tree_height( *node );
}

template < typename NodeContent, typename Tag >
constexpr uint64_t get_memory_usage( const oc_proxy_node< NodeContent, Tag >& node )
{
        return get_memory_usage( *node );
}

template < typename Tag >
constexpr std::array< point< 3, Tag >, 8 > cube_offsets{ { point< 3, Tag >{ 1, 1, 1 },
                                                           point< 3, Tag >{ -1, 1, 1 },
                                                           point< 3, Tag >{ 1, -1, 1 },
                                                           point< 3, Tag >{ -1, -1, 1 },
                                                           point< 3, Tag >{ 1, 1, -1 },
                                                           point< 3, Tag >{ -1, 1, -1 },
                                                           point< 3, Tag >{ 1, -1, -1 },
                                                           point< 3, Tag >{ -1, -1, -1 } } };

template < typename NodeContent, typename Tag >
constexpr std::array< point< 3, Tag >, 8 >
get_node_corners( const oc_proxy_node< NodeContent, Tag >& node )
{
        return em::map_f_to_a( cube_offsets< Tag >, [&]( const point< 3, Tag >& p ) {
                return p * float( *node.get_edge_length() / 2 ) +
                       cast_to_vec( node.get_position() );
        } );
}

constexpr std::array< vec< 3 >, 3 > cube_intersection_axis{ {
    vec< 3 >( 0, 0, 1 ),
    vec< 3 >( 0, 1, 0 ),
    vec< 3 >( 1, 0, 0 ),
} };

template < typename NodeContent, typename RealTag, typename NodeTag = RealTag >
class oc_node_full_collision_query : i_collision_item< const point< 3, RealTag >*, const vec< 3 >* >
{
      public:
        using tag                      = RealTag;
        using point_iterator           = const point< 3, RealTag >*;
        using separation_axis_iterator = const vec< 3 >*;

      private:
        const oc_proxy_node< NodeContent, NodeTag > node_;
        const pose< RealTag >                       offset_;

        std::array< point< 3, RealTag >, 8 > points_;
        std::array< vec< 3 >, 3 >            intersection_axes_;

      public:
        constexpr explicit oc_node_full_collision_query(
            oc_proxy_node< NodeContent, RealTag > node )
          : node_( std::move( node ) )
          , offset_()
          , points_( get_node_corners( node_ ) )
          , intersection_axes_( cube_intersection_axis )
        {
        }

        constexpr oc_node_full_collision_query( oc_proxy_node< NodeContent, NodeTag > node,
                                                pose< RealTag >                       offset )
          : node_( std::move( node ) )
          , offset_( std::move( offset ) )
        {

                points_ = em::map_f_to_a( get_node_corners( node_ ),
                                          [&]( const point< 3, NodeTag >& p ) {  //
                                                  return transform( p, offset_ );
                                          } );

                intersection_axes_ = em::map_f_to_a( cube_intersection_axis,
                                                     [&]( const vec< 3 >& v ) {  //
                                                             return transform( v, offset_ );
                                                     } );
        }

        constexpr em::view< separation_axis_iterator > get_separation_axes() const final
        {
                return { nullptr, nullptr };
        }

        constexpr em::view< separation_axis_iterator > get_intersection_axes() const final
        {
                return em::view{ intersection_axes_ };
        }

        constexpr em::view< point_iterator > get_points() const final
        {
                return em::view{ points_ };
        }

        constexpr bool is_leaf() const
        {
                return node_.is_leaf();
        }

        constexpr auto get_children() const
        {
                using query = oc_node_full_collision_query< NodeContent, RealTag, NodeTag >;
                return std::array< query, 8 >{ query( node_.get_child( 0 ), offset_ ),  //
                                               query( node_.get_child( 1 ), offset_ ),
                                               query( node_.get_child( 2 ), offset_ ),
                                               query( node_.get_child( 3 ), offset_ ),
                                               query( node_.get_child( 4 ), offset_ ),
                                               query( node_.get_child( 5 ), offset_ ),
                                               query( node_.get_child( 6 ), offset_ ),
                                               query( node_.get_child( 7 ), offset_ ) };
        }

        constexpr const oc_proxy_node< NodeContent, NodeTag >& operator*() const
        {
                return node_;
        }

        constexpr const oc_proxy_node< NodeContent, NodeTag >* operator->() const
        {
                return &node_;
        }
};

template < typename NodeContent, typename RealTag, typename NodeTag = RealTag >
class oc_node_fast_collision_query
  : i_fast_collision_item< oc_node_full_collision_query< NodeContent, RealTag, NodeTag >, 3 >
{
      public:
        using tag       = RealTag;
        using full_type = oc_node_full_collision_query< NodeContent, RealTag, NodeTag >;

      private:
        const point< 3, RealTag >                   pos_;
        const em::radius                            rad_;
        const oc_proxy_node< NodeContent, NodeTag > node_;
        const pose< RealTag >                       offset_;

      public:
        explicit oc_node_fast_collision_query( oc_proxy_node< NodeContent, RealTag > node )
          : oc_node_fast_collision_query( std::move( node ), pose< tag >{} )
        {
        }
        explicit oc_node_fast_collision_query( oc_proxy_node< NodeContent, NodeTag > node,
                                               pose< RealTag >                       offset )
          : pos_( transform( node.get_position(), offset ) )
          , rad_( node.get_edge_length() * std::sqrt( 2.f ) / 2.f )
          , node_( std::move( node ) )
          , offset_( std::move( offset ) )
        {
        }

        constexpr point< 3, RealTag > get_min_sphere_position() const final
        {
                return pos_;
        }

        constexpr em::radius get_min_sphere_radius() const final
        {
                return rad_;
        }

        constexpr full_type get_full() const final
        {
                return full_type{ node_, offset_ };
        }

        constexpr bool is_leaf() const
        {
                return node_.is_leaf();
        }

        constexpr auto get_children() const
        {
                using query = oc_node_fast_collision_query< NodeContent, RealTag, NodeTag >;
                return std::array< query, 8 >{ query( node_.get_child( 0 ), offset_ ),  //
                                               query( node_.get_child( 1 ), offset_ ),
                                               query( node_.get_child( 2 ), offset_ ),
                                               query( node_.get_child( 3 ), offset_ ),
                                               query( node_.get_child( 4 ), offset_ ),
                                               query( node_.get_child( 5 ), offset_ ),
                                               query( node_.get_child( 6 ), offset_ ),
                                               query( node_.get_child( 7 ), offset_ ) };
        }

        constexpr const oc_proxy_node< NodeContent, NodeTag >& operator*() const
        {
                return node_;
        }

        constexpr const oc_proxy_node< NodeContent, NodeTag >* operator->() const
        {
                return &node_;
        }
};

template < typename LhContent,
           typename RhContent,
           typename RealTag,
           typename LhNodeTag,
           typename RhNodeTag >
constexpr bool operator>( const oc_node_full_collision_query< LhContent, RealTag, LhNodeTag >& lh,
                          const oc_node_full_collision_query< RhContent, RealTag, RhNodeTag >& rh )
{
        return lh->get_edge_length() > rh->get_edge_length();
}
template < typename LhContent,
           typename RhContent,
           typename RealTag,
           typename LhNodeTag,
           typename RhNodeTag >
constexpr bool operator>( const oc_node_fast_collision_query< LhContent, RealTag, LhNodeTag >& lh,
                          const oc_node_fast_collision_query< RhContent, RealTag, RhNodeTag >& rh )
{
        return lh->get_edge_length() > rh->get_edge_length();
}

template < collision_mode Mode, typename NodeContent, typename RealTag, typename NodeTag = RealTag >
using oc_node_collision_query =
    std::conditional_t< Mode == collision_mode::FAST,
                        oc_node_fast_collision_query< NodeContent, RealTag, NodeTag >,
                        oc_node_full_collision_query< NodeContent, RealTag, NodeTag > >;

template < typename NodeContent, typename Tag, typename UnaryFunction >
inline std::optional< em::min_max< point< 3, Tag > > >
get_used_space( const oc_proxy_node< NodeContent, Tag >& node, UnaryFunction f )
{
        if ( f( node ) ) {
                return { dimensional_min_max_elem< point< 3, Tag > >( get_node_corners( node ) ) };
        }
        if ( !node.has_children() ) {
                return {};
        }

        std::vector< point< 3, Tag > > v;
        for ( const oc_proxy_node< NodeContent, Tag >& child : node.generate_children() ) {
                auto opt_subres = get_used_space( child, f );
                if ( !opt_subres ) {
                        continue;
                }
                v.push_back( opt_subres->min );
                v.push_back( opt_subres->max );
        }
        if ( v.empty() ) {
                return {};
        }
        return { dimensional_min_max_elem< point< 3, Tag > >( v ) };
}

template < typename NodeContent, typename Tag, typename UnaryFunction >
constexpr em::volume get_volume( const oc_proxy_node< NodeContent, Tag >& node,
                                 UnaryFunction                            accept_f )
{
        if ( accept_f( node ) ) {
                return em::volume( float( std::pow( *node.get_edge_length(), 3 ) ) );
        }
        if ( !node.has_children() ) {
                return em::volume( 0 );
        }
        return em::sum( node.generate_children(),
                        [&]( const oc_proxy_node< NodeContent, Tag >& child ) {  //
                                return getVolume( child, accept_f );
                        } );
}

template < typename NodeContent, typename Tag >
inline bool validate_tree_children_depth( const oc_proxy_node< NodeContent, Tag >& node )
{
        if ( node.get_depth() == node.get_max_depth() ) {
                if ( node.has_children() ) {
                        SCHPIN_INFO_LOG( "octree.proxy_node", "Node has children in max depth" );
                }
                return !node.has_children();
        }

        if ( !node.has_children() ) {
                return true;
        }

        return em::all_of( node.generate_children(),
                           []( const oc_proxy_node< NodeContent, Tag >& child ) {  //
                                   return validate_tree_children_depth( child );
                           } );
}

template < typename NodeContent, typename Tag >
constexpr oc_proxy_node< NodeContent, Tag >
find_leaf_node( const oc_proxy_node< NodeContent, Tag >& node, const point< 3, Tag >& p )
{
        if ( !node.has_children() ) {
                return node;
        }
        auto children = node.generate_children();
        auto child_iter =
            em::find_if( children, [&]( const oc_proxy_node< NodeContent, Tag >& child ) {  //
                    return contains( child, p );
            } );

        return find_leaf_node( *child_iter, p );
}
}  // namespace schpin
