// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "base.h"
#include "base/physics.h"
#include "node.h"
#include "octree/interfaces.h"
#include "proxy_node.h"

#include <utility>

#pragma once

namespace schpin
{
template < typename NodeContent, typename Tag >
class oc_tree
{
        oc_node< NodeContent > root_;
        oc_node_context< Tag > root_context_;
        NodeContent            default_content_;

      public:
        using node    = oc_node< NodeContent >;
        using content = NodeContent;

        oc_tree() = default;

        oc_tree( const NodeContent& default_content, em::length root_length )
          : root_( default_content )
          , root_context_{ 0U, 0U, root_length, point< 3, Tag >( 0, 0, 0 ) }
          , default_content_( default_content )
        {
        }

        oc_tree( oc_node< NodeContent >        root,
                 const oc_node_context< Tag >& context,
                 NodeContent                   default_content )
          : root_( std::move( root ) )
          , root_context_( context )
          , default_content_( std::move( default_content ) )
        {
                // data that we got may be completely wrong, wrong data-relative-to-childs in nodes,
                // the depth of root thing... I propose a static factory function, that is only one
                // with access to this constructor It shal return datatype with option for error and
                // force user to handle that TODO
                SCHPIN_ASSERT( context.depth == 0 );  // this bothers me and makes me feel insecure
                oc_passive_visitor< NodeContent, Tag > visitor{};
                this->modify( visitor );
        }

        oc_tree( const oc_tree& other ) = delete;
        oc_tree& operator=( const oc_tree& other ) = delete;

        oc_tree( oc_tree&& other ) noexcept = default;
        oc_tree& operator=( oc_tree&& other ) noexcept = default;

        oc_tree copy() const
        {
                return oc_tree( root_.copy(), root_context_, default_content_ );
        }

        // again, this is wrong because the proxy root is not const
        oc_proxy_node< NodeContent, Tag > get_proxy_root() const
        {
                // NOLINTNEXTLINE(cppcoreguidelines-pro-type-const-cast)
                return { const_cast< oc_node< NodeContent >& >( root_ ), root_context_ };
        }

        const oc_node< NodeContent >& get_bare_root() const
        {
                return root_;
        }

        template < typename Visitor >
        void modify( const Visitor& visitor )
        {
                static_assert(
                    std::is_base_of< i_oc_simple_visitor< NodeContent, Tag >, Visitor >::value,
                    "This method accepts only datatype implementing IOcSimpleVisitor" );
                oc_simple_visitor_overlay< Visitor, NodeContent, Tag > overlay( visitor );
                modify( oc_empty_info_type{}, overlay );
        }

        template < typename InfoType, typename Visitor >
        void modify( const InfoType& info, const Visitor& visitor )
        {
                static_assert(
                    std::is_base_of< i_oc_visitor< InfoType, NodeContent, Tag >, Visitor >::value,
                    "This method accepts only datatype implementing IOcVisitor" );
                oc_proxy_node< NodeContent, Tag > n{ root_, root_context_ };
                modify_node( n, info, visitor );
                SCHPIN_ASSERT( validate_octree( *this ) );
        }

        template < typename Container >
        void expand( const Container& cont )
        {
                for ( const auto& p : cont ) {
                        expand( p );
                }
        }

        void expand( const point< 3, Tag >& target )
        {
                if ( contains( get_proxy_root(), target ) ) {
                        return;
                }
                // The actual root cube has to be expanded in direction of that target.
                // The center of new root cube has to exists in direction of the tagert point.
                //
                // To solve this, the relative coordinates of actual root cube, in frame of the new
                // root, have to be known.

                point< 3, Tag > root_pos   = root_context_.position;
                oc_child_index child_index = get_oc_child_index_based_on_center( target, root_pos );

                // New root for this tree is created and actual root is set as it's child at
                // appropiate position

                std::array< oc_node< NodeContent >, 8 > new_nodes( {
                    oc_node< NodeContent >( default_content_ ),
                    oc_node< NodeContent >( default_content_ ),
                    oc_node< NodeContent >( default_content_ ),
                    oc_node< NodeContent >( default_content_ ),
                    oc_node< NodeContent >( default_content_ ),
                    oc_node< NodeContent >( default_content_ ),
                    oc_node< NodeContent >( default_content_ ),
                    oc_node< NodeContent >( default_content_ ),
                } );
                std::swap( new_nodes.at( *child_index ), root_ );
                root_ = oc_node< NodeContent >(
                    std::make_unique< std::array< oc_node< NodeContent >, 8 > >(
                        std::move( new_nodes ) ),
                    default_content_ );

                root_->child_based_update( root_ );

                // Scale the new properties of the root cube.
                root_context_ = upscale_context( root_context_, child_index );

                expand( target );
        }

        const oc_node_context< Tag >& get_root_context() const
        {
                return root_context_;
        }

        const NodeContent& get_default_content() const
        {
                return default_content_;
        }

        ~oc_tree() = default;

      private:
        template < typename InfoType, typename Visitor >
        void modify_node( oc_proxy_node< NodeContent, Tag >& node,
                          const InfoType&                    parent_info,
                          const Visitor&                     visitor )
        {
                const oc_proxy_node< NodeContent, Tag >& c_node = node;
                InfoType                                 info;
                if ( node.get_depth() == node.get_max_depth() ) {
                        node.modify_content( [&]( NodeContent& content ) {  //
                                visitor.access_bottom_leaf( c_node, parent_info, content );
                        } );
                        return;
                }
                if ( node.has_children() ) {
                        std::pair< oc_node_control, InfoType > res =
                            visitor.access_node( c_node, parent_info );
                        if ( res.first == oc_node_control::END ) {
                                return;
                        }
                        info = res.second;
                } else {
                        std::pair< oc_locked_leaf_control, InfoType > res =
                            visitor.access_locked_leaf( c_node, parent_info );
                        if ( res.first == oc_locked_leaf_control::UNLOCK ) {
                                node.modify_content( [&]( NodeContent& content ) {
                                        visitor.access_bottom_leaf( c_node, parent_info, content );
                                } );
                                return;
                        }
                        if ( res.first == oc_locked_leaf_control::END ) {
                                return;
                        }
                        node.init_children();
                        info = res.second;
                }
                if ( node.has_children() ) {
                        for ( oc_proxy_node< NodeContent, Tag > child : node.generate_children() ) {
                                modify_node( child, info, visitor );
                        }
                }

                node.modify_content( [&]( NodeContent& content ) {  //
                        content.child_based_update( *node );
                } );
                node.try_prunning();
        }
};

template < typename NodeContent, typename Tag >
constexpr bool operator==( const oc_tree< NodeContent, Tag >& lh,
                           const oc_tree< NodeContent, Tag >& rh )
{
        return lh.get_bare_root() == rh.get_bare_root() &&
               lh.get_root_context() == rh.get_root_context();
}

template < typename NodeContent, typename Tag, typename UnaryFunction >
inline std::optional< em::min_max< point< 3, Tag > > >
get_used_space( const oc_tree< NodeContent, Tag >& node, UnaryFunction&& f )
{
        return get_used_space( node.get_proxy_root(), std::forward< UnaryFunction >( f ) );
}

template < typename NodeContent, typename Tag, typename UnaryFunction >
void log_octree_space( const oc_tree< NodeContent, Tag >& tree, UnaryFunction f )
{
        em::min_max< point< 3, Tag > > res = get_used_space( tree.get_proxy_root(), f );
        SCHPIN_DEBUG_LOG( "octree", "Octree min: " << res.min << " max: " << res.max );
}

template < typename NodeContent, typename Tag >
std::optional< oc_proxy_node< NodeContent, Tag > >
find_leaf_node( const oc_tree< NodeContent, Tag >& tree, const point< 3, Tag >& p )
{
        if ( !contains( tree.get_proxy_root(), p ) ) {
                return {};
        }
        return { find_leaf_node( tree.get_proxy_root(), p ) };
}

template < typename NodeContent, typename Tag >
inline bool validate_octree( const oc_tree< NodeContent, Tag >& tree )
{
        if ( !validate_tree_children_depth( tree.get_proxy_root() ) ) {
                return false;
        }

        if ( get_tree_height( tree.get_proxy_root() ) != tree.get_root_context().max_depth + 1 ) {
                SCHPIN_INFO_LOG( "tree",
                                 "tree with height: " << get_tree_height( tree.get_proxy_root() )
                                                      << " and maxdepth: "
                                                      << tree.get_root_context().max_depth );
                return false;
        }

        return true;
}

}  // namespace schpin

template < typename NodeContent, typename Tag >
struct std::hash< schpin::oc_tree< NodeContent, Tag > >
{
        std::size_t operator()( const schpin::oc_tree< NodeContent, Tag >& tree )
        {
                return std::hash< schpin::oc_node< NodeContent > >()( tree.get_bare_root() ) ^
                       std::hash< schpin::oc_node_context< Tag > >()( tree.get_root_context() );
        }
};
